/*
lynx.c - Atari Lynx support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "lynx.h"
//#include "lynxer.h" 

#include "../backup/unknown_bu.h"
#include "../backup/unknown_bu512.h"

// LNX is a format for Lynx ROMs used by some emulators
#define lnx_HEADER_START 0
#define lnx_HEADER_LEN 64

typedef struct
{
  unsigned char   magic[4];
  unsigned short   page_size_bank0;
  unsigned short   page_size_bank1;
  unsigned short   version;
  unsigned char   cartname[32];
  unsigned char   manufname[16];
  unsigned char   rotation;
  unsigned char   spare[5];
}LYNX_HEADER_NEW;
static LYNX_HEADER_NEW header;

int lynx_lyx(struct ucon64_ *rom)
{
	char buf[MAXBUFSIZE];

	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	strcpy(buf,rom->rom);
	newext(buf,".LYX");
	
	filecopy(rom->rom,rom->buheader_len,quickftell(rom->rom),filebackup(buf),"wb");

	printf("Wrote output to %s\n", rom->rom);
	return(0);	
}

/*
int lynx_test(struct ucon64_ *rom)
{
  lynxer_main(rom);
	return(0);
}
*/

int lynx_lnx(struct ucon64_ *rom)
{
	char buf[MAXBUFSIZE];

	if(rom->header_len!=0)
	{
		printf("ERROR: this seems to be already an LNX file; use -nhd to override\n\n");
		return(-1);
	}

	header.page_size_bank0=0;
	header.page_size_bank1=0;

//	if( !( ( quickftell( rom->rom )/256 )%512 ) )
	{
		header.page_size_bank0=(
				(quickftell(rom->rom)>(4*MBIT)) ?
				( (4*MBIT) /256 ) :
				( quickftell(rom->rom)/256 ) 
			);
		header.page_size_bank1=(quickftell(rom->rom)-(4*MBIT))/256;
	}
//	else
	{
		printf(
"ERROR: the ROM does not have the correct length use -b0 and -b1 to set the\n\
       header values manually\n\n");
	}

	memset(header.cartname,0,sizeof(header.cartname));
	memset(header.manufname,0,sizeof(header.manufname));
	memset(header.spare,0,sizeof(header.spare));

	header.version=1;
	header.magic[0]='L';
	header.magic[1]='Y';
	header.magic[2]='N';
	header.magic[3]='X';

	header.rotation=0;

	strncpy(header.cartname,rom->rom,sizeof(header.cartname));
	strcpy(header.manufname,"Atari");

	strcpy(buf,rom->rom);
	newext(buf,".LNX");

	quickfwrite(&header,0,sizeof(header),filebackup(buf),"wb");

	filecopy(rom->rom,0,quickftell(rom->rom),buf,"ab");
	
	printf("Wrote output to %s\n", rom->rom);
	return(0);
}

int lynx_nrot(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	header.rotation=0; //no rotation

	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");
	
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");
	
	return(0);
}

int lynx_rotl(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	header.rotation=1; //rotate left	

	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");
	
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");
	
	return(0);
}

int lynx_rotr(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	header.rotation=2; //rotate right

	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");
	
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");
	
	return(0);
}

int lynx_n(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	memset(header.cartname,0,sizeof(header.cartname));
	strncpy(header.cartname,rom->file,sizeof(header.cartname));

	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");

	return(0);
}

int lynx_b0(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	if(	(atol(rom->file)%64)!=0 ||
		(atol(rom->file)>512)
	)header.page_size_bank0=0;
	else header.page_size_bank0=atol(rom->file)*4;

	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");
	
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	return(0);
}

int lynx_b1(struct ucon64_ *rom)
{
	if(!rom->header_len)
	{
		printf("ERROR: this is no LNX file; use -hd to override\n\n");
		return(-1);
	}

	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	quickfread(&header,0,sizeof(LYNX_HEADER_NEW),rom->rom);

	if(	(atol(rom->file)%64)!=0 ||
		(atol(rom->file)>512)
	)header.page_size_bank1=0;
	else header.page_size_bank1=atol(rom->file)*4;
	quickfwrite(&header,0,sizeof(LYNX_HEADER_NEW),filebackup(rom->rom),"r+b");
	
	filehexdump(rom->rom,0,sizeof(LYNX_HEADER_NEW));
	printf("\n");

	return(0);
}

int lynx_init(struct ucon64_ *rom)
{
typedef struct
{
  unsigned char   magic[4];
  unsigned short   page_size_bank0;
  unsigned short   page_size_bank1;
  unsigned short   version;
  unsigned char   cartname[32];
  unsigned char   manufname[16];
  unsigned char   rotation;
  unsigned char   spare[5];
}LYNX_HEADER_NEW;
LYNX_HEADER_NEW header;


rom->console=ucon64_LYNX;

rom->buheader_len=0;
ucon64_dbsearch(rom);

if(rom->db_crc32==0)
{
/*  rom->buheader_len=unknown_bu512_HEADER_LEN;
  rom->current_crc32=fileCRC32(rom->rom,rom->buheader_len);

  ucon64_dbsearch(rom);

*/  if(rom->db_crc32==0)
  {
    quickfread(&header,0,sizeof(header),rom->rom);
  
    if(!strncmp(header.magic,"LYNX",4))
    {
      rom->buheader_len=0;
    }
    else if(!argcmp(rom->argc,rom->argv,"-lynx")) 
    {
      rom->console=ucon64_UNKNOWN;
      return(-1);
    }
  }
}

  if(argcmp(rom->argc,rom->argv,"-hd"))rom->buheader_len=unknown_bu512_HEADER_LEN;
  if(argcmp(rom->argc,rom->argv,"-nhd"))rom->buheader_len=0;

  strcpy(rom->title,lynx_TITLE);
        
if(!strncmp(header.magic,"LYNX",4))
{
  rom->buheader_len=0;
  rom->header_start=0;
  rom->header_len=lnx_HEADER_LEN;
  quickfread(rom->header,0,sizeof(header),rom->rom);        

  strcpy(rom->name,header.cartname);
  strcpy(rom->manufacturer,header.manufname);
                

  sprintf(rom->misc,
    "Internal Size: Bank0 %ld Bytes (%.4f Mb); Bank1 %ld Bytes (%.4f Mb)\n"
    "Version: %d\n"
    "Rotation: %s"
    ,(long)header.page_size_bank0*256
    ,((float)(header.page_size_bank0*256)/MBIT)
    ,(long)header.page_size_bank1*256
    ,((float)(header.page_size_bank1*256)/MBIT)
    ,header.version
    ,(!header.rotation) ? "No" : ( ( header.rotation==1 ) ? "Left" : "Right" )
  );

}
else
{
  strcpy(rom->misc,"");
  rom->header_start=0;
  rom->header_len=0;
  quickfread(&header,0,sizeof(header),rom->rom);
}

  strcpy(rom->copier,(!rom->buheader_len) ? unknown_bu_TITLE : unknown_bu512_TITLE );

  return(0);
}

int lynx_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -lynx		force recognition"
#ifndef DB
"; NEEDED"
#endif

	"\n  -hd		force ROM has LNX header (+64 Bytes)\n"
	"  -nhd		force ROM has no LNX header\n"
	"  *		show info (default)\n"
//	"  -o2lyx        convert *.O file to LYX format (for developers)\n"
	"  -lyx		convert to LYX/RAW (strip 64 Bytes LNX header)\n"
	"  -lnx		convert to LNX (uses default values for the header);\n"
	"		adjust the LNX header with the following options\n"
	"  -n		change ROM name (LNX only); $FILE=NEWNAME\n"
	"  -nrot		set no rotation (LNX only)\n"
	"  -rotl		set rotation left (LNX only)\n"
	"  -rotr		set rotation right (LNX only)\n"
	"  -b0		change Bank0 kBytes size (LNX only); $FILE={0,64,128,256,512}\n"
	"  -b1		change Bank1 kBytes size (LNX only); $FILE={0,64,128,256,512}\n"
,lynx_TITLE
);

printf("\n");

return(0);
}

