/*
n64.c - Nintendo64 support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "n64.h"

#include "../patch/ips.h"
#include "../patch/aps.h"

#include "../backup/doctor64.h"
#include "../backup/doctor64jr.h"
#include "../backup/cd64.h"

#include "../backup/unknown_bu512.h"

#define nintendo64_ROM_HEADER_START 0x0
#define nintendo64_ROM_HEADER_LEN 64

struct nintendo64_chksum_
{
  unsigned long crc1;
  unsigned long crc2;
} nintendo64_chksum__;

long x;
char sram[512];
char buf[32768];
char *usmsbuf;

#ifdef BACKUP
int nintendo64_xv64(struct ucon64_ *rom)
{
if(!access(rom->rom,F_OK))
{
  if(doctor64_write(	rom->rom
				,rom->buheader_len
				,quickftell(rom->rom)
				,rom->parport	)!=0)
  {
     printf(
"ERROR:  please check cables and connection\n"
"	turn off/on the backup unit\n"
"	splitted ROMs must be joined first\n"
"	use $FILE={3bc,378,278,...} to specify your port\n"
"	set the port to SPP (Standard, Normal) mode in your bios\n"
     );

        return(-1);
    }
}
else
{
	if(doctor64_read(	rom->rom
				,rom->parport	)!=0)
	{
		printf(
"ERROR:  please check cables and connection\n"
"	turn off/on the backup unit\n"
"	splitted ROMs must be joined first\n"
"	use $FILE={3bc,378,278,...} to specify your port\n"
"	set the port to SPP (Standard, Normal) mode in your bios\n"
);
		return(-1);
	}
  }
  printf("\n");
  return(0);
}

int nintendo64_xdjr(struct ucon64_ *rom)
{
if(!access(rom->rom,F_OK))
{
    if(doctor64jr_write(	rom->rom
				,rom->buheader_len
				,quickftell(rom->rom)
				,rom->parport	)!=0)
    {

      printf(
"ERROR:  please check cables and connection\n"
"	turn off/on the backup unit\n"
"	splitted ROMs must be joined first\n"
"	use $FILE={3bc,378,278,...} to specify your port\n"
"	set the port to SPP (Standard, Normal) mode in your bios\n"
      );

      return(-1);
    }
  }
  else
  {
	if(doctor64jr_read(	rom->rom
				,rom->parport	)!=0)
    {
      printf(
"ERROR:  please check cables and connection\n"
"	turn off/on the backup unit\n"
"	splitted ROMs must be joined first\n"
"	use $FILE={3bc,378,278,...} to specify your port\n"
"	set the port to SPP (Standard, Normal) mode in your bios\n"
      );

      return(-1);
    }
  }
  printf("\n");
  return(0);
}
#endif

int nintendo64_sram(struct ucon64_ *rom)
{
//	This is the support of LaC's makesram routine for uploading
//	SRAM files to a Cart SRAM.  The .v64 file is his work, not mine

	if(quickftell(rom->file)!=512 ||
	quickftell(rom->rom)!=1310720)
	{
		printf("ERROR: check if ROM has 1310720 Bytes and SRAM has 512 Bytes\n");
		printf("       or the ROM is too short to calculate checksum\n");
		return(-1);
	}

	quickfread(sram,0,sizeof(sram),rom->file);
	        
	if(rom->interleaved!=0)strswap(sram,0,sizeof(sram));

	quickfwrite(sram,0x0286C0,sizeof(sram),filebackup(rom->rom),"r+b");

	return(0);
}

int nintendo64_v64(struct ucon64_ *rom)
{
	if(rom->interleaved!=0)
	{
		printf("ERROR: Already in V64 format (use -nswp to override)\n");
		return(-1);
	}
//	strcpy(rom->rom,fmkrom(rom->rom,rom->buheader_len));
	strcpy(buf,rom->rom);
	newext(buf,".V64");

	filecopy(rom->rom,0,quickftell(rom->rom),filebackup(buf),"wb");
	fileswap(buf,0,quickftell(buf));
	printf("Wrote output to %s\n", buf);
	return(0);
}

int nintendo64_n(struct ucon64_ *rom)
{
	strcpy(buf,rom->file);
//	strcat(option2,"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0");
        	strcat(buf,"                     ");
	if(rom->interleaved!=0)strswap(buf,0,strlen(buf));
	quickfwrite(buf,nintendo64_ROM_HEADER_START+rom->buheader_len+32,20,filebackup(rom->rom),"r+b");
	return(0);
}

int nintendo64_f(struct ucon64_ *rom)
{
//TODO pal/ntsc fix
  return 0;
}

int nintendo64_chk(struct ucon64_ *rom)
{
  filehexdump(rom->rom,0x10+rom->buheader_len,8);
filebackup(rom->rom);
  nintendo64_chksum(rom);

  for(x=0;x<4;x++)
  {
    quickfputc( rom->rom
     ,nintendo64_ROM_HEADER_START+rom->buheader_len+0x10+(x^rom->interleaved)
,(nintendo64_chksum__.crc1&0xff000000)>>24
     ,"r+b"
     );
    nintendo64_chksum__.crc1<<=8;
  }

  for(x=0;x<4;x++)
  {
    quickfputc( rom->rom
     ,nintendo64_ROM_HEADER_START+rom->buheader_len+0x14+(x^rom->interleaved)
,(nintendo64_chksum__.crc2&0xff000000)>>24
     ,"r+b"
     );
    nintendo64_chksum__.crc2<<=8;
  }

  filehexdump(rom->rom,0x10+rom->buheader_len,8);

  return(0);
}

int nintendo64_z64(struct ucon64_ *rom)
{
	if(!rom->interleaved)
	{
		printf("ERROR: Already in Z64 format (use -swp to override)\n");
		return(-1);
	}

	strcpy(buf,rom->rom);
	newext(buf,".Z64");

	filecopy(rom->rom,0,quickftell(rom->rom),filebackup(buf),"wb");
	fileswap(buf,0,quickftell(buf));

  printf("Wrote output to %s\n", buf);
  return 0;
}

int nintendo64_bot(struct ucon64_ *rom)
{
	if(!access(rom->file,F_OK))
	{
		quickfread(	buf
				,0
				,((sizeof(buf)<quickftell(rom->file)) ?
				sizeof(buf) :
				quickftell(rom->file))
				,rom->file
			);
			
		if(rom->interleaved!=0)strswap(buf,0,sizeof(buf));
		strhexdump(buf,0,0,sizeof(buf));

		quickfwrite(
				buf
				,nintendo64_ROM_HEADER_START
				+rom->buheader_len
				+0x040
				,0x01000-0x040
				,filebackup(rom->rom)
				,"r+b"
			);
	}
	else
	{
		strcpy(buf,rom->rom);
		newext(buf,".BOT");

		filecopy(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+0x040,0x01000-0x040,filebackup(buf),"wb");

		if(rom->interleaved!=0)fileswap(buf,0,quickftell(buf));
	}
	return(0);
}


int nintendo64_p(struct ucon64_ *rom)
{
	filepad(filebackup(rom->rom),rom->buheader_len,MBIT);
	return 0;
}

int nintendo64_swap(struct ucon64_ *rom)
{
/*
	strcpy(buf,rom->rom);
	newext(buf,".ROM");

	fcopy(rom->rom,0,buf,"wb");
//	if(!rom->interleaved)fileswap(buf,0,fsize(buf,rom->buheader_len));
	fileswap(buf,0,fsize(buf,rom->buheader_len));
*/
	return 0;
}

int nintendo64_usms(struct ucon64_ *rom)
{
//must be smaller than 4mbit 524288 bytes will be inserted
//from 1b410 to 9b40f (7ffff)

	if(!access(rom->file,F_OK))
	{
		if(quickftell(rom->file)>0x07fffff)return(-1);
		if(!(usmsbuf=(char *)malloc((quickftell(rom->file)+2)*sizeof(char))))return(-1);
		quickfread(	rom->file
				,0
				,quickftell(rom->file)
				,usmsbuf
			);

		if(rom->interleaved!=0)strswap(usmsbuf,0,quickftell(rom->file));

		filecopy(rom->rom,0,quickftell(rom->rom),filebackup("patched.v64"),"wb");
		quickfwrite(	usmsbuf
				,nintendo64_ROM_HEADER_START
				+rom->buheader_len
				+0x01b410
				,quickftell(rom->file)
				,"patched.v64"
				,"r+b"
			);
		free(usmsbuf);
		strcpy(rom->rom,"patched.v64");
	}
	else
	{
		strcpy(buf,rom->rom);
		newext(buf,".GG");
		filecopy(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+0x040,0x01000-0x040,filebackup(buf),"wb");
		if(rom->interleaved!=0)fileswap(buf,0,quickftell(buf));
	}
	return(0);
}


int nintendo64_init(struct ucon64_ *rom)
{
long x=0,y=0;
char buf[4096];

char *nintendo64_MANUFACTURER[]={
"0 Unknown Manufacturer",
"1",
"2",
"3",
"4",
"5",
"6",
"7",
"8",
"9",
"10",
"11",
"12",
"13",
"14",
"15",
"16",
"17",
"18",
"19",
"20",
"21",
"22",
"23",
"24",
"25",
"26",
"27",
"28",
"29",
"30",
"31",
"32",
"33",
"34",
"35",
"36",
"37",
"38",
"39",
"40",
"41",
"42",
"43",
"44",
"45",
"46",
"47",
"48",
"49",
"50",
"51",
"52",
"53",
"54",
"55",
"56",
"57",
"58",
"59",
"60",
"61",
"62",
"63",
"64",
"65",
"66",
"Nintendo",
"68",
"69",
"70",
"71",
"72",
"73",
"74",
"75",
"76",
"77",
"Nintendo",
"79",
"80",
"81",
"82",
"83",
"84",
"85",
"86",
"87",
"88",
"89",
"90",
"91",
"92",
"93",
"94",
"95",
"96",
"97",
"98",
"99",
"100",
"101",
"102",
"103",
"104",
"105",
"106",
"107",
"108",
"109",
"110",
"111",
"112",
"113",
"114",
"115",
"116",
"117",
"118",
"119",
"120",
"121",
"122",
"123",
"124",
"125",
"126",
"127",
"128",
"129",
"130",
"131",
"132",
"133",
"134",
"135",
"136",
"137",
"138",
"139",
"140",
"141",
"142",
"143",
"144",
"145",
"146",
"147",
"148",
"149",
"150",
"151",
"152",
"153",
"154",
"155",
"156",
"157",
"158",
"159",
"160",
"161",
"162",
"163",
"164",
"165",
"166",
"167",
"168",
"169",
"170",
"171",
"172",
"173",
"174",
"175",
"176",
"177",
"178",
"179",
"180",
"181",
"182",
"183",
"184",
"185",
"186",
"187",
"188",
"189",
"190",
"191",
"192",
"193",
"194",
"195",
"196",
"197",
"198",
"199",
"200",
"201",
"202",
"203",
"204",
"205",
"206",
"207",
"208",
"209",
"210",
"211",
"212",
"213",
"214",
"215",
"216",
"217",
"218",
"219",
"220",
"221",
"222",
"223",
"224",
"225",
"226",
"227",
"228",
"229",
"230",
"231",
"232",
"233",
"234",
"235",
"236",
"237",
"238",
"239",
"240",
"241",
"242",
"243",
"244",
"245",
"246",
"247",
"248",
"249",
"250",
"251",
"252",
"253",
"254",
"255"
};

char *nintendo64_COUNTRY[]={\
"0 Unknown Country",
"1",
"2",
"3",
"4",
"5",
"6",
"7",
"8",
"9",
"10",
"11",
"12",
"13",
"14",
"15",
"16",
"17",
"18",
"19",
"20",
"21",
"22",
"23",
"24",
"25",
"26",
"27",
"28",
"29",
"30",
"31",
"32",
"33",
"34",
"35",
"36",
"37",
"38",
"39",
"40",
"41",
"42",
"43",
"44",
"45",
"46",
"47",
"48",
"49",
"50",
"51",
"52",
"53",
"54",
"55",
"56",
"57",
"58",
"59",
"60",
"61",
"62",
"63",
"64",
"65",
"66",
"67",
"Germany",
"U.S.A.",
"70",
"71",
"72",
"73",
"Japan",
"75",
"76",
"77",
"78",
"79",
"Europe",
"81",
"82",
"83",
"84",
"Australia",
"86",
"87",
"France, German, Dutch",
"89",
"90",
"91",
"92",
"93",
"94",
"95",
"96",
"97",
"98",
"99",
"100",
"101",
"102",
"103",
"104",
"105",
"106",
"107",
"108",
"109",
"110",
"111",
"112",
"113",
"114",
"115",
"116",
"117",
"118",
"119",
"120",
"121",
"122",
"123",
"124",
"125",
"126",
"127",
"128",
"129",
"130",
"131",
"132",
"133",
"134",
"135",
"136",
"137",
"138",
"139",
"140",
"141",
"142",
"143",
"144",
"145",
"146",
"147",
"148",
"149",
"150",
"151",
"152",
"153",
"154",
"155",
"156",
"157",
"158",
"159",
"160",
"161",
"162",
"163",
"164",
"165",
"166",
"167",
"168",
"169",
"170",
"171",
"172",
"173",
"174",
"175",
"176",
"177",
"178",
"179",
"180",
"181",
"182",
"183",
"184",
"185",
"186",
"187",
"188",
"189",
"190",
"191",
"192",
"193",
"194",
"195",
"196",
"197",
"198",
"199",
"200",
"201",
"202",
"203",
"204",
"205",
"206",
"207",
"208",
"209",
"210",
"211",
"212",
"213",
"214",
"215",
"216",
"217",
"218",
"219",
"220",
"221",
"222",
"223",
"224",
"225",
"226",
"227",
"228",
"229",
"230",
"231",
"232",
"233",
"234",
"235",
"236",
"237",
"238",
"239",
"240",
"241",
"242",
"243",
"244",
"245",
"246",
"247",
"248",
"249",
"250",
"251",
"252",
"253",
"254",
"255"
};

rom->console=ucon64_N64;

rom->buheader_len=0;
x+=quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len)*256*256*256;
x+=quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+1)*256*256;
x+=quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+2)*256;
x+=quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+3);

rom->interleaved=rom->buheader_len=0;
if(x!=0x80371240)
{
	rom->interleaved=1;rom->buheader_len=0;
	if(x!=0x37804012)
	{
		if(!argcmp(rom->argc,rom->argv,"-n64"))
		{
                        rom->console=ucon64_UNKNOWN;
                        return(-1);
		}
	}
}

if(argcmp(rom->argc,rom->argv,"-hd"))rom->buheader_len=unknown_bu512_HEADER_LEN;
if(argcmp(rom->argc,rom->argv,"-nhd"))rom->buheader_len=0;

if(argcmp(rom->argc,rom->argv,"-nswp"))rom->interleaved=0;
if(argcmp(rom->argc,rom->argv,"-swp"))rom->interleaved=1;

/*
#ifdef TR_ENDIAN_IS_LITTLE
        rominfo.validation           = *(tr_u16 *)(rom.header + (0x00 ^ 0x02));
        rominfo.compression          = *(tr_u8 *) (rom.header + (0x02 ^ 0x03));
        rominfo.unknown1             = *(tr_u8 *) (rom.header + (0x03 ^ 0x03));
        rominfo.clockrate            = *(tr_u32 *) (rom.header +  0x04);
        rominfo.programcounter       = *(tr_u32 *) (rom.header +  0x08);
        rominfo.release              = *(tr_u32 *) (rom.header +  0x0c);
        rominfo.crc1                 = *(tr_u32 *) (rom.header +  0x10);
        rominfo.crc2                 = *(tr_u32 *) (rom.header +  0x14);
        rominfo.unknown2_h           = *(tr_u32 *) (rom.header +  0x18);
        rominfo.unknown2_l           = *(tr_u32 *) (rom.header +  0x1c);
        for(i=0; i<20; i++)
                rominfo.name[i^0x03] = rom.header[i+0x20];
        rominfo.unknown3             = *(tr_u8 *) (rom.header + (0x34 ^ 0x03));
        rominfo.unknown4             = *(tr_u8 *) (rom.header + (0x35 ^ 0x03));
        rominfo.unknown5             = *(tr_u8 *) (rom.header + (0x36 ^ 0x03));
        rominfo.unknown6             = *(tr_u8 *) (rom.header + (0x37 ^ 0x03));
        rominfo.unknown7             = *(tr_u8 *) (rom.header + (0x38 ^ 0x03));
        rominfo.unknown8             = *(tr_u8 *) (rom.header + (0x39 ^ 0x03));
        rominfo.unknown9             = *(tr_u8 *) (rom.header + (0x3a ^ 0x03));
        rominfo.manufacturerid       = *(tr_u8 *) (rom.header + (0x3b ^ 0x03));
        rominfo.cartridgeid          = *(tr_u16 *)(rom.header + (0x3c ^ 0x02));
        rominfo.countrycode          = *(tr_u8 *) (rom.header + (0x3e ^ 0x03));
        rominfo.unknown10            = *(tr_u8 *) (rom.header + (0x3f ^ 0x03));
#endif
   
        rom.prgcode_base_orig = rominfo.programcounter;
        rom.prgcode_base = rom.prgcode_base_orig & 0x1fffffff;

        printf("ROM length      : 0x%08lx (%lu) Bytes\n", rom.length, rom.length);
        printf("ROM valid. hword: %04hx\n",     rominfo.validation);
        printf("ROM comp. byte  : %01x (%s)\n",
                rominfo.compression,
                rominfo.compression == 0x12
                        ? "uncompressed"
                        : rominfo.compression == 0x13
                                ? "compressed"
                                : "?");
        printf("ROM ????????????: %01x\n",      rominfo.unknown1);
        printf("ROM clock rate  : %08"PR32"x\n",     rominfo.clockrate);
        printf("ROM pc          : %08"PR32"x\n",     rominfo.programcounter);
        printf("ROM release     : %08"PR32"x\n",     rominfo.release);
        printf("ROM crc 1       : %08"PR32"x\n",     rominfo.crc1);
        printf("ROM crc 2       : %08"PR32"x\n",     rominfo.crc2);
        printf("ROM name        : \"%s\"\n",      rominfo.name);
        printf("ROM manufac. id : %02x (%s)\n", rominfo.manufacturerid, manufacturerid[rominfo.manufacturerid]);
        printf("ROM cartridgeid : %04hx\n",     rominfo.cartridgeid);
        printf("ROM country code: %02x (%s)\n", rominfo.countrycode, countrycode[rominfo.countrycode]);
        puts("-------------------------------------------------------------------------------"); fflush(stdout);

*/

  rom->header_start=nintendo64_ROM_HEADER_START;
  rom->header_len=nintendo64_ROM_HEADER_LEN;
  quickfread(rom->header,rom->header_start,rom->header_len,rom->rom);

  strcpy(rom->title,nintendo64_TITLE);

if(!rom->buheader_len)
{
  if(!rom->interleaved)
    strcpy(rom->copier,"Z64");
  else
    sprintf(rom->copier,"%s\n%s\n%s",doctor64_TITLE,doctor64jr_TITLE,cd64_TITLE );
}
else strcpy(rom->copier,unknown_bu512_TITLE);

  quickfread(rom->name,nintendo64_ROM_HEADER_START+32+rom->buheader_len,20,rom->rom);
  rom->name[20]=0;
  if(rom->interleaved!=0)strswap(rom->name,0,strlen(rom->name));
  rom->name[20]=0;

  strcpy(rom->manufacturer,nintendo64_MANUFACTURER[quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+((!rom->interleaved)?59:58))]);

  strcpy(rom->country,nintendo64_COUNTRY[quickfgetc(rom->rom,nintendo64_ROM_HEADER_START+rom->buheader_len+((!rom->interleaved)?62:63))]);

  strcpy(rom->misc,"");

  rom->has_internal_crc=1;
  rom->internal_crc_len=rom->internal_crc2_len=4;

  nintendo64_chksum(rom);

  rom->current_internal_crc=nintendo64_chksum__.crc1;

  for(x=0;x<4;x++)
  {
    rom->internal_crc<<=8;
    rom->internal_crc+=quickfgetc( rom->rom
     ,nintendo64_ROM_HEADER_START+rom->buheader_len+0x10+(x^rom->interleaved));
  }

  for(x=0;x<4;x++)
  {
    y<<=8;
    y+=quickfgetc( rom->rom
     ,nintendo64_ROM_HEADER_START+rom->buheader_len+0x14+(x^rom->interleaved));
  }

    sprintf(buf,"2nd Checksum: %%s, 0x%%0%dlx (calculated) %%s= 0x%%0%dlx (internal)"
             ,rom->internal_crc2_len*2,rom->internal_crc2_len*2);

    sprintf(rom->internal_crc2,buf
           ,(nintendo64_chksum__.crc2 == y) ? "ok" : "bad (use -chk to fix)",
           nintendo64_chksum__.crc2,
           (nintendo64_chksum__.crc2 == y) ? "=" : "!",
           y);

 return(0);
}

#define CHECKSUM_START 0x1000
#define CHECKSUM_LENGTH 0x100000L
#define CHECKSUM_HEADERPOS 0x10
#define CHECKSUM_END (CHECKSUM_START + CHECKSUM_LENGTH)
#define CHECKSUM_STARTVALUE 0xf8ca4ddc
#define ROL(i, b) (((i)<<(b)) | ((i)>>(32-(b))))
#define max2(a, b) ( (a)>(b) ? (a) : (b) )
#define min2(a, b) ( (a)<(b) ? (a) : (b) )
#define BYTES2LONG(b, s) ( (((b)[0^(s)] & 0xffL) << 24) | \
                           (((b)[1^(s)] & 0xffL) << 16) | \
                           (((b)[2^(s)] & 0xffL) <<  8) | \
                           (((b)[3^(s)] & 0xffL)) )

#define LONG2BYTES(l, b, s)  (b)[0^(s)] = ((l)>>24)&0xff; \
                             (b)[1^(s)] = ((l)>>16)&0xff; \
                             (b)[2^(s)] = ((l)>> 8)&0xff; \
                             (b)[3^(s)] = ((l)    )&0xff;

#define HEADER_MAGIC 0x80371240

int nintendo64_chksum(struct ucon64_ *rom)
{
//   ROM Checksum Routine Courtesy of:
//   chksum64 V1.2, a program to calculate the ROM checksum of Nintendo64 ROMs.
//   Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

char readchunk3[32768];
    unsigned long i;
    unsigned long c1, k1, k2;
    unsigned long t1, t2, t3, t4;
    unsigned long t5, t6;
    unsigned int n;
    long clen = CHECKSUM_LENGTH;
    long rlen;
	FILE *fh;
	unsigned long sum1, sum2;
    rlen = (quickftell(rom->rom)-rom->buheader_len) - CHECKSUM_START;
    t1 = CHECKSUM_STARTVALUE;
    t2 = CHECKSUM_STARTVALUE;
    t3 = CHECKSUM_STARTVALUE;
    t4 = CHECKSUM_STARTVALUE;
    t5 = CHECKSUM_STARTVALUE;
    t6 = CHECKSUM_STARTVALUE;

	if((quickftell(rom->rom)-rom->buheader_len)<0x0100000)//0x0101000
		return(-1);//rom is too short

	if(!(fh=fopen(rom->rom,"r+b")))return(-1);
	fseek(fh, CHECKSUM_START+rom->buheader_len, SEEK_SET);

for( ;; )
{
	if( rlen > 0 )
	{
        	n = fread(readchunk3, 1, min2(sizeof(readchunk3), clen), fh);
        	if( (n & 0x03) != 0 )
        	{
          		n += fread(readchunk3+n, 1, 4-(n&3), fh);
        	}
      	}
      	else
      	{
        	n = min2(sizeof(readchunk3), clen);
      	}
      	if( (n == 0) || ((n&3) != 0) )
      	{
        	if( (clen != 0) || (n != 0) )
        	{
          		fprintf(stderr, "ERROR: short read, checksum may be incorrect.\n");
        	}
        	break;
      	}
      	for( i=0; i<n; i+=4 )
      	{
	        c1 = BYTES2LONG(&readchunk3[i], rom->interleaved);
	        k1 = t6 + c1;
	        if( k1 < t6 ) t4++;
	        t6 = k1;
	        t3 ^= c1;
	        k2 = c1 & 0x1f;
	        k1 = ROL(c1, k2);
	        t5 += k1;
	        if( c1 < t2 )
	        {
	        	t2 ^= k1;
	        }
	        else
	        {
	        	t2 ^= t6 ^ c1;
	        }
	        t1 += c1 ^ t5;
      	}
      	if( rlen > 0 )
      	{
        	rlen -= n;
        	if( rlen <= 0 ) memset(readchunk3, 0, sizeof(readchunk3));
      	}
	clen -= n;
}
sum1 = t6 ^ t4 ^ t3;
sum2 = t5 ^ t2 ^ t1;

  LONG2BYTES(sum1, &readchunk3[0], 0);
  LONG2BYTES(sum2, &readchunk3[4], 0);
  fseek(fh, CHECKSUM_HEADERPOS, SEEK_SET);
  fread(readchunk3+8, 1, 8, fh);
  fseek(fh, CHECKSUM_HEADERPOS, SEEK_SET);
  LONG2BYTES(sum1, &readchunk3[16], rom->interleaved);
  LONG2BYTES(sum2, &readchunk3[20], rom->interleaved);
  fclose(fh);


nintendo64_chksum__.crc1=sum1;
nintendo64_chksum__.crc2=sum2;

	return(0);
	
}


int nintendo64_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -n64          force recognition\n"
	"  -swp          force ROM is swapped (2143)\n"
	"  -nswp         force ROM is not swapped (1234)\n"
	"  *             show info (default)\n"
	"  -n            change ROM name; $FILE=NEWNAME\n"
	"  -v64          convert to Doctor V64 (and compatibles/swapped)\n"
	"  -z64          convert to Z64 (Zip Drive/not swapped)\n"
//	"TODO:  -f       remove NTSC/PAL protection\n"
	"  -bot          add/extract boot code to(from) ROM; $FILE=BOOTCODE\n"
	"                extracts automatically when $FILE does not exist (4032 Bytes)\n"
	"                supports: 6101, 6102, 6103, 6105 and 6106 boot codes\n"
	"  -sram         LAC's Makesram; $ROM=(LAC's SRAM ROM image) $FILE=SRAMFILE\n"
	"                the SRAMFILE must have a size of 512 Bytes\n"
	"  -usms         Jos Kwanten's ultraSMS (Sega Master System/GameGear Emulator);\n"
	"                $ROM=(Jos Kwanten's ultraSMS ROM image) $FILE=SMSROM(<=4Mb)\n"
	"  -chk          fix ROM checksum\n"
//	"TODO:  -bios    enable backup in Doctor V64 BIOS; $ROM=BIOS\n"
	"  -p            pad ROM to full Mb\n"
//	"TODO:  -gs      apply GameShark code (permanent); $FILE=CODE\n"
,nintendo64_TITLE
);

aps_usage(argc,argv);
ips_usage(argc,argv);

#ifdef BACKUP
doctor64_usage(argc,argv);
doctor64jr_usage(argc,argv);
cd64_usage(argc,argv);
#endif

printf("\n");


return(0);
}

