/*
neogeo.c - NeoGeo support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "neogeo.h"

#include "../backup/unknown_bu.h"
#include "../backup/unknown_bu512.h"

#include "../backup/cdrw.h"

#define neogeo_ROM_HEADER_START 0x0
#define neogeo_ROM_HEADER_LEN 0

char buf[MAXBUFSIZE];

int neogeo_bios(struct ucon64_ *rom)
{
	strcpy(buf,rom->rom);
	newext(buf,".NEW");

	filecopy(rom->rom,0,MBIT,filebackup(buf),"wb");
	fileswap(buf,0,MBIT);

	return(0);
}

int neogeo_sam(struct ucon64_ *rom)
{
	if(sam2wav(rom->rom)==-1)
	{
		printf("ERROR: SAM header seems to be corrupt\n");
	}
	return(0);
}

int neogeo_mgd(struct ucon64_ *rom)
{
	if(rom->splitted[0]!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}
	return(0);
}

int neogeo_mvs(struct ucon64_ *rom)
{
	if(rom->splitted[0]!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}
	return(0);
}

int neogeo_s(struct ucon64_ *rom)
{
	if(rom->splitted[0]!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}
	return(0);
}


int neogeo_init(struct ucon64_ *rom)
{
rom->console=ucon64_NEOGEO;

rom->buheader_len=0;
ucon64_dbsearch(rom);

if(rom->db_crc32==0)
{
/*	rom->buheader_len=unknown_bu512_HEADER_LEN;
	rom->current_crc32=fileCRC32(rom->rom,rom->buheader_len);

	ucon64_dbsearch(rom);

	if(rom->db_crc32==0)
*/	{
		if(!argcmp(rom->argc,rom->argv,"-ng"))
		{
                        rom->console=ucon64_UNKNOWN;
                        return(-1);
		}
	}
}

  if(argcmp(rom->argc,rom->argv,"-hd"))rom->buheader_len=unknown_bu512_HEADER_LEN;
  if(argcmp(rom->argc,rom->argv,"-nhd"))rom->buheader_len=0;

  strcpy(rom->title,neogeo_TITLE);
  strcpy(rom->copier,(!rom->buheader_len) ? unknown_bu_TITLE : unknown_bu512_TITLE );

  return(0);
}

int sam2wav(char *filename)
{
	unsigned char buf[32];
	FILE *fh, *fh2;
	unsigned datasize, wavesize, riffsize, freq, bits, rate;

	if(quickftell(filename)<16)return(-1);
	if(!(fh = fopen(filename, "rb")))return(-1);

	strcpy(buf,filename);
	buf[findlast(buf,".")+1]=0;
	strcat(buf,findlwr(buf)?"wav":"WAV");
                                                
	if(!(fh2 = fopen(buf, "wb")))return(-1);
	fread(buf, 1, 4, fh);

	if(strncmp(buf, "MAME", 4) != 0 )return(-1);
	fread(buf, 1, 4, fh);

	datasize = buf[0] + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24);
	wavesize = /* "fmt " */ 4 + 4 + 4 + 2 + 2;
	riffsize = /* RIFF */ 4 + 4 + 4 + 4 + 4 + wavesize + datasize;

	buf[0] = riffsize & 0xff;
	buf[1] = (riffsize >> 8) & 0xff;
	buf[2] = (riffsize >> 16) & 0xff;
	buf[3] = (riffsize >> 24) & 0xff;

	fwrite("RIFF", 1, 4, fh2);
	fwrite(buf, 1, 4, fh2);

	fwrite("WAVE", 1, 4, fh2);
	fwrite("fmt ", 1, 4, fh2);

	buf[0] = wavesize & 0xff;
	buf[1] = (wavesize >> 8) & 0xff;
	buf[2] = (wavesize >> 16) & 0xff;
	buf[3] = (wavesize >> 24) & 0xff;
	fwrite(buf, 1, 4, fh2);

	/* number of channels - alway 1 for MAME samples */
	fwrite("\001\000\001\000", 1, 4, fh2);

	/* read frequency */
	fread(buf, 1, 4, fh);
	freq = buf[0] + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24);

	/* write frequency */
	fwrite(buf, 1, 4, fh2);

	/* read bits, amplitude and extension */
	fread(buf, 1, 4, fh);

	bits = buf[0];

	if(bits == 16)rate = freq * 2;
	else rate = freq;

	buf[0] = rate & 0xff;
	buf[1] = (rate >> 8) & 0xff;
	buf[2] = (rate >> 16) & 0xff;
	buf[3] = (rate >> 24) & 0xff;

	/* write bytes per second */
	fwrite(buf, 1, 4, fh2);	

	/* write alignment */
	if( bits == 16 )fwrite("\002\000", 1, 2, fh2);
	else fwrite("\001\000", 1, 2, fh2);

	buf[0] = bits & 0xff;
	buf[1] = (bits >> 8) & 0xff;

	/* write bits per sample */
	fwrite(buf, 1, 2, fh2);
	fwrite("data", 1, 4, fh2);

	buf[0] = datasize & 0xff;
	buf[1] = (datasize >> 8) & 0xff;
	buf[2] = (datasize >> 16) & 0xff;
	buf[3] = (datasize >> 24) & 0xff;
	fwrite(buf, 1, 4, fh2);

	if( bits == 16 )
	{
		for( ; ; )
		{
			if( fread(buf, 1, 2, fh) != 2 )break;
			fwrite(buf, 1, 2, fh2);
		}
	}
	else
	{
	for( ; ; )
        {
			if( fread(buf, 1, 1, fh) != 1 )break;
			buf[0] = buf[0] ^ 0x80;
			fwrite(buf, 1, 1, fh2);
        }
    }
	fclose(fh2);
	fclose(fh);
	return( 0);
}


int neogeo_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -ng		force recognition"
#ifndef DB
"; NEEDED"
#endif

	"\n  -ns		force ROM is not splitted\n"
	"  *		show info (default)\n"
//	"TODO:  -mgd	convert to Multi Game Doctor/MGD2/RAW\n"
//	"TODO:  -mvs	convert to Arcade/MVS\n"
	"  -bios		convert NeoCd Bios to work with NeoCD emulator; $ROM=BIOS\n"
	"		http://www.illusion-city.com/neo/\n"
//	"TODO:  -j	join splitted ROM\n"
//	"TODO:  -s	split ROM into 4Mb parts (for backup unit(s) with fdd)\n"
//	"TODO:  -ngs	convert Neo Geo sound to WAV; $ROM=*_m1.rom or *_v*.rom\n"
	"  -sam		convert SAM/M.A.M.E. sound to WAV; $ROM=SAMFILE\n"
//	"TODO:  -chkm	check/fix Multiple Arcade Machine Emulator/M.A.M.E. ROMs;\n"
//	"		$ROM=DIRECTORY\n"
,neogeo_TITLE
);

#ifdef BACKUP
#endif

#ifdef CD
  cdrw_raw_usage(argc,argv);
#endif

printf("\n");

return(0);
}

#ifdef CD
  int neogeo_rawt(struct ucon64_ *rom)
  {
    return( cdrw_raw_mktoc(rom) );
  }

  int neogeo_xraw(struct ucon64_ *rom)
  {
    return( (!access(rom->rom, F_OK)) ? cdrw_raw_write(rom) : cdrw_raw_read(rom) );
  }
#endif
