/*
nes.c - Nintendo Entertainment System support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "nes.h"

#include "../backup/unknown_bu.h"
#include "../backup/unknown_bu512.h"

int nes_splitted=0;

#define nes_ROM_HEADER_START 0x0
#define nes_ROM_HEADER_LEN 16

long x,y;
char header[512];
//int nes_argc;
//char *nes_argv[16];
 char name[4];
  int J,NewGame;

int nes_gge(struct ucon64_ *rom)
{
	char buf[32768];

	if(!gameGenieEncodeNES(rom->rom, buf))
	{
		printf("%-12s = %s\n", rom->rom, buf);
	}
	else
	{
		printf("ERROR: %-12s is badly formed\n", rom->rom);
	}
	return(0);
}

int nes_ggd(struct ucon64_ *rom)
{
	char buf[32768];

	if(!gameGenieDecodeNES(rom->rom, buf))
	{
		printf("%-12s = %s\n", rom->rom, buf);
	}
	else
	{
		printf("ERROR: %-12s is badly formed\n", rom->rom);
	}
	return(0);
}

int nes_j(struct ucon64_ *rom)
{
	return(nes_join(rom->rom));
}


int nes_gg(struct ucon64_ *rom)
{
	char buf[32768];

	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	if(gameGenieDecodeNES(rom->file, buf)!=0)
	{
		printf("ERROR: %-12s is badly formed\n", rom->file);
		return(-1);
	}
	printf("%-12s = %s\n", rom->file, buf);

	sscanf(buf,"%lx:%lx:*",&x,&y);

	printf("\n");
	filehexdump(rom->rom,x+rom->buheader_len,1);

	buf[0]=(char)y;
	quickfwrite(buf,x+rom->buheader_len,1,filebackup(rom->rom),"r+b");

	filehexdump(rom->rom,x+rom->buheader_len,1);
	printf("\n");
	return(0);
}


int nes_ffe(struct ucon64_ *rom)
{
	char buf[32768];

	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	strcpy(buf,filebackup(rom->rom));
	newext(buf,".FFE");

	memset(header,0,sizeof(header));
	header[1]=(((quickftell(rom->rom)-rom->buheader_len)/8192)/256);
	header[0]=(((quickftell(rom->rom)-rom->buheader_len)/8192)-(header[1]*256));
	                
	quickfwrite(header,0,sizeof(header),filebackup(buf),"wb");
        
	filecopy(rom->rom,rom->buheader_len,quickftell(rom->rom),buf,"ab");
	printf("Wrote output to %s\n", rom->rom);
	return(0);
}

int nes_fds(struct ucon64_ *rom)
{
	char buf[32768], *buf2;

	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	strcpy(buf,rom->rom);
	newext(buf,".FDS");

	if((quickftell(rom->rom)-262192)>0 ||
		!(buf2=(unsigned char *)malloc(66000)))return(-1);

	for(J=0;J<4;J++)
	{
		if(quickfread(buf2,0,65500,rom->rom)!=65500)break;

// Check if game is the same
		if(J)NewGame=memcmp(name,buf2+16,4);
		else
		{
			memcpy(name,buf2+16,4);
			NewGame=0;
		}

// Check disk image for validity
		if(buf2[0]!=1 || buf2[56]!=2 ||	buf2[58]!=3) break;
		if(!NewGame&&(buf2[21]!=(J&1))) break;

// Write a disk image
		if(quickfwrite(buf2,1,65500,buf,"ab")!=65500)break;
	}

	free(buf2);
	return(0);
}

int nes_fdsl(struct ucon64_ *rom)
{
	return(nes_fdslist(rom->rom));
}

int nes_ines(struct ucon64_ *rom)
{
	char buf[32768];

	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	strcpy(buf,filebackup(rom->rom));
	newext(buf,".NES");

	filecopy(rom->rom,rom->buheader_len,quickftell(rom->rom),buf,"wb");
	return(0);
}


int nes_ineshd(struct ucon64_ *rom)
{
	char buf[32768];

	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	strcpy(buf,filebackup(rom->rom));
	newext(buf,".HDR");

	filecopy(rom->rom,rom->buheader_len,16,filebackup(buf),"wb");
	return(0);
}

int nes_pas(struct ucon64_ *rom)
{
	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	return(0);
}

int nes_s(struct ucon64_ *rom)
{
	if(nes_splitted!=0)
	{
		printf("ERROR: this ROM is seems to be splitted (use -ns to override)\n");
		return(-1);
	}

	return(0);
}

int nes_init(struct ucon64_ *rom)
{
long x = 0;
char buf[4096];

rom->console=ucon64_NES;

rom->buheader_len=0;
ucon64_dbsearch(rom);

if(rom->db_crc32==0)
{
/*	rom->buheader_len=unknown_bu512_HEADER_LEN;
	rom->current_crc32=fileCRC32(rom->rom,rom->buheader_len);

	ucon64_dbsearch(rom);

	if(rom->db_crc32==0)
*/	{
rom->buheader_len=0;
quickfread(buf,0,5,rom->rom);
if (	buf[0]!='N' ||
	buf[1]!='E' ||
	buf[2]!='S' /*||
	buf[4]!=1 ||
	buf[5]!=1*/
)
{
	rom->buheader_len=unknown_bu512_HEADER_LEN;
	quickfread(buf,rom->buheader_len,5,rom->rom);
	if (	buf[0]!='N' ||
		buf[1]!='E' ||
		buf[2]!='S' ||
		buf[4]!=1/* ||
		buf[5]!=1*/
	)
	{
		if(!argcmp(rom->argc,rom->argv,"-nes"))
		{
                        rom->console=ucon64_UNKNOWN;
			return(-1);
		}
	}
}
	}
}

/*
strcpy(buf,rom.rom);
if(!strcmp(strupr(&buf[strlen(buf)-4]),".FDS")&&(quickftell(rom.rom)%65500)==0)
	rom.console=ucon64_NES;


*/


  if(argcmp(rom->argc,rom->argv,"-hd"))rom->buheader_len=unknown_bu512_HEADER_LEN;
  if(argcmp(rom->argc,rom->argv,"-nhd"))rom->buheader_len=0;

  nes_splitted=testsplit(rom->rom);
  if(argcmp(rom->argc,rom->argv,"-ns"))nes_splitted=0;

  strcpy(rom->title,nes_TITLE);
  strcpy(rom->copier,(!rom->buheader_len) ? unknown_bu_TITLE : unknown_bu512_TITLE );

  rom->header_start=nes_ROM_HEADER_START;
  rom->header_len=nes_ROM_HEADER_LEN;
  quickfread(rom->header,rom->header_start,rom->header_len,rom->rom);
  strcpy(rom->misc,"");

  sprintf(buf,"Internal Size: %.4f Mb\n",((float) ((int)quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+4)<<14)/MBIT)+((float)((int) quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+5)<<13)/MBIT));
  strcat(rom->misc,buf);
  
  sprintf(buf,"Internal PRG Size: %.4f Mb\n",(float) ((int)quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+4)<<14)/MBIT);
  strcat(rom->misc,buf);
  sprintf(buf,"Internal CHR/RAM Size: %.4f Mb\n",(float)((int) quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+5)<<13)/MBIT);
  strcat(rom->misc,buf);
  
  x=quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+6);
  sprintf(buf,"Memory mapper: %d\n",(int) ((x >> 4) | (quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+7) & 0xf0)));
  strcat(rom->misc,buf);
  
  sprintf(buf,"Mirroring: %s\n",(x& 0x01)?"vertical":"horizontal");
  strcat(rom->misc,buf);
  sprintf(buf,"Save RAM: %s\n",(x& 0x02)?"yes":"no");
  strcat(rom->misc,buf);
  sprintf(buf,"512-byte trainer: %s\n",(x& 0x04)?"yes":"no");
  strcat(rom->misc,buf);
  sprintf(buf,"Four-screen mirroring: %s\n",(x& 0x08)?"yes":"no");
  strcat(rom->misc,buf);
  sprintf(buf,"VSSystem: %s",(quickfgetc(rom->rom,nes_ROM_HEADER_START+rom->buheader_len+7)& 0x01)?"yes":"no");
  strcat(rom->misc,buf);
  
  return(0);
}


int nes_join(char *name)
{
char *buf3;
char buf[4096],buf2[4096];
char nesname[4096];
char romheader[16];

//detect multifile pasofami format and join them into one single .NES file

if(!(buf3=strrchr(name,'.')))buf3=name;
else buf3++;
strcpy(buf,buf3);
strupr(buf);

if(!strdcmp(buf,"HDR")||\
!strdcmp(buf,"PRM")||\
!strdcmp(buf,"700")||\
!strdcmp(buf,"CHR")||\
!strdcmp(buf,"PRG"))
{
	strcpy(nesname,name);
	newext(nesname,".NES");

	strcpy(buf,name);
	newext(buf,".PRM");

	quickfread(buf2,0,quickftell(buf),buf);

	sprintf(romheader,"NES%c%c%c%c%c%c%c%c%c%c%c%c%c",0,0,0,0,0,0,0,0,0,0,0,0,0);
	quickfwrite(romheader,0,nes_ROM_HEADER_LEN,nesname,"wb");

	strcpy(buf,name);

	buf[findlast(buf,".")+1]=0;
	newext(buf,".PRG");

	filecopy(buf,0,quickftell(buf),nesname,"ab");

	buf[findlast(buf,".")+1]=0;
	newext(buf,".CHR");

	filecopy(buf,0,quickftell(buf),nesname,"ab");

	strcpy(name,nesname);
}
if(!strdcmp(buf,"FDS"))
{
//fds >> nes
}
return(0);
}







#define encodeNES(v, n, m, s) data[n] |= (v >> s) & m
#define decodeNES(v, n, m, s) v |= (data[n] & m) << s

char mapNesChar(char hex) {
  switch(toupper(hex)) {
    case '0': return 'A';
    case '1': return 'P';
    case '2': return 'Z';
    case '3': return 'L';
    case '4': return 'G';
    case '5': return 'I';
    case '6': return 'T';
    case '7': return 'Y';
    case '8': return 'E';
    case '9': return 'O';
    case 'A': return 'X';
    case 'B': return 'U';
    case 'C': return 'K';
    case 'D': return 'S';
    case 'E': return 'V';
    case 'F': return 'N';
    default: break;
  }
  return '?';
}

char unmapNesChar(char genie) {
  switch(toupper(genie)) {
    case 'A': return '0';
    case 'P': return '1';
    case 'Z': return '2';
    case 'L': return '3';
    case 'G': return '4';
    case 'I': return '5';
    case 'T': return '6';
    case 'Y': return '7';
    case 'E': return '8';
    case 'O': return '9';
    case 'X': return 'A';
    case 'U': return 'B';
    case 'K': return 'C';
    case 'S': return 'D';
    case 'V': return 'E';
    case 'N': return 'F';
    default: break;
  }
  return '0';
}

int
isNesChar(char c) {
  return strchr("APZLGITYEOXUKSVN-", toupper(c)) != 0;
}

int
gameGenieDecodeNES(const char *in, char *out) {
  int address = 0;
  int value = 0;
  int check = 0;
  int haveCheck = 0;
  int data[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
  int i;

  if (strlen(in) == 8) {
    haveCheck = 1;
  } else if (strlen(in) == 6) {
    haveCheck = 0;
  } else {
    return -1;
  }
  for (i=0; i<strlen(in); ++i) if (!isNesChar(in[i])) return -1;

  for (i=0; i<6; ++i) data[i] = hexValue(unmapNesChar(in[i]));
  if (haveCheck) {
    data[6] = hexValue(unmapNesChar(in[6]));
    data[7] = hexValue(unmapNesChar(in[7]));
  } 

  address = 0x8000; /* force high bit on */
  decodeNES(address, 1, 8, 4);
  decodeNES(address, 2, 7, 4);
  decodeNES(address, 3, 7, 12);
  decodeNES(address, 3, 8, 0);
  decodeNES(address, 4, 7, 0);
  decodeNES(address, 4, 8, 8);
  decodeNES(address, 5, 7, 8);
  if (haveCheck) {
    decodeNES(value, 0, 7, 0);
    decodeNES(value, 0, 8, 4);
    decodeNES(value, 1, 7, 4);
    decodeNES(value, 7, 8, 0);
    decodeNES(check, 6, 7, 0);
    decodeNES(check, 6, 8, 0);
    decodeNES(check, 6, 8, 4);
    decodeNES(check, 7, 7, 4);
  } else {
    decodeNES(value, 0, 7, 0);
    decodeNES(value, 0, 8, 4);
    decodeNES(value, 1, 7, 4);
    decodeNES(value, 5, 8, 0);
  }

  if (haveCheck) {
    sprintf(out, "%04X:%02X:%02X", address, value, check);
  } else {
    sprintf(out, "%04X:%02X", address, value);
  }

  return 0;
}

int
gameGenieEncodeNES(const char *in, char *out) {
  int address = 0x8000;
  int value = 0;
  int check = 0;
  int haveCheck = 0;
  int data[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
  int i;

  if (strlen(in) == 10 && in[4] == ':' && in[7] == ':') {
    haveCheck = 1;
  } else if (strlen(in) == 7 && in[4] == ':') {
    haveCheck = 0;
  } else {
    return -1;
  }
  for (i=0; i<strlen(in); ++i)
    if (in[i] != ':' && !isxdigit((int) in[i]))
      return -1;

  if (haveCheck) {
    sscanf(in, "%x:%x:%x", &address, &value, &check);
  } else {
    sscanf(in, "%x:%x", &address, &value);
  }

  /* Encode address with transposition cipher. 
   * Do not encode the high address bit (optional, Galoob isn't
   * consistent but usually doesn't.
   */
  encodeNES(address, 1, 8, 4);
  encodeNES(address, 2, 7, 4);
  encodeNES(address, 3, 7, 12);
  encodeNES(address, 3, 8, 0);
  encodeNES(address, 4, 7, 0);
  encodeNES(address, 4, 8, 8);
  encodeNES(address, 5, 7, 8);
  if (haveCheck) {
    encodeNES(value, 0, 7, 0);
    encodeNES(value, 0, 8, 4);
    encodeNES(value, 1, 7, 4);
    encodeNES(value, 7, 8, 0);
    encodeNES(check, 6, 7, 0);
    encodeNES(check, 6, 8, 0);
    encodeNES(check, 6, 8, 4);
    encodeNES(check, 7, 7, 4);
  } else {
    encodeNES(value, 0, 7, 0);
    encodeNES(value, 0, 8, 4);
    encodeNES(value, 1, 7, 4);
    encodeNES(value, 5, 8, 0);
  }

  if (haveCheck) {
    sprintf(out, "%c%c%c%c%c%c%c%c", 
	    mapNesChar(hexDigit(data[0])),
	    mapNesChar(hexDigit(data[1])),
	    mapNesChar(hexDigit(data[2])),
	    mapNesChar(hexDigit(data[3])),
	    mapNesChar(hexDigit(data[4])),
	    mapNesChar(hexDigit(data[5])),
	    mapNesChar(hexDigit(data[6])),
	    mapNesChar(hexDigit(data[7])));
  } else {
    sprintf(out, "%c%c%c%c%c%c", 
	    mapNesChar(hexDigit(data[0])),
	    mapNesChar(hexDigit(data[1])),
	    mapNesChar(hexDigit(data[2])),
	    mapNesChar(hexDigit(data[3])),
	    mapNesChar(hexDigit(data[4])),
	    mapNesChar(hexDigit(data[5])));
  }

  return 0;
}






int nes_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -nes		force recognition\n"
	"  *		show info/contents (default); $ROM=ROM or $ROM=DISKIMAGE\n"
	"  -hd		force ROM has FFE header (+512 Bytes)\n"
	"  -nhd		force ROM has no FFE header\n"
	"  -ns		force ROM is not splitted\n"
	"  -ffe		convert to FFE (+512 Bytes)\n"
	"  -ines		convert to iNES(Emu)/RAW\n"
	"  -ineshd	extract iNES header from ROM (16 Bytes)\n"
//	"TODO:  -fam	convert to FAM(jap. Emu)\n"
//	"  -fds		convert FAM to Famicon DiskSystem/FDS (diskimage)\n"
//	"  -fdsl		list Famicon DiskSystem/FDS (diskimage) contents\n"
	"  -j		join Pasofami/PRM/PRG/CHR/700/splitted ROM\n"
//	"TODO:  -pas,-s	convert/split to Pasofami/PRM/PRG/CHR/700 (Emulator)\n"
//	"TODO:  -tr	truncate doubled PRG/CHR/RAM\n"
//	"TODO:  -nfs	convert NFS sound to WAV; $ROM=NFSFILE\n"
	"  -gge		encode GameGenie code; $ROM=AAAA:VV or $ROM=AAAA:VV:CC\n"
	"  -ggd		decode GameGenie code; $ROM=XXXXXX or $ROM=XXXXXXXX\n"
	"  -gg		apply GameGenie code (permanent);\n"
	"		$FILE=XXXXXX or $FILE=XXXXXXXX\n"
	"\n"
,nes_TITLE
);

return(0);
}


int nes_fdslist(char *nes_diskimage)
{
FILE *fh;
char buf[32768];
unsigned char DiskHeader[] =
  {
    0x01,0x2A,0x4E,0x49,0x4E,0x54,0x45,0x4E,
    0x44,0x4F,0x2D,0x48,0x56,0x43,0x2A
  };
  char S[16],T[16];
  int Disk,Files,Start,Size,Total,J,I;
  /* Open the disk image */
  if(!(fh=fopen(nes_diskimage,"rb")))
  {
    return(-1);
  }

  /* Find total disk image size */
  fseek(fh,0,SEEK_END);
  Total=ftell(fh)/65500;
  J=ftell(fh)%65500;

  /* Check out if it is integer number of disks */
  if(J) printf("%d excessive bytes\n",J);

  /* Scan through disks */
  rewind(fh);
  for(Disk=0;Disk<Total;Disk++)
  {
    /* Seek to the next disk */
    fseek(fh,Disk*65500,SEEK_SET);

    /* Read the disk header */
    if(fread(buf,1,58,fh)!=58)
    {
      printf("Can't read disk header\n");
      return(-1);
    }

    /* Check if disk header ID s valid */
    if(memcmp(buf,DiskHeader,15))
    {
      printf("Invalid disk header\n");
      return(-1);
    }

    /* Check if file number header ID is valid */
    if(buf[56]!=2)
    {
      printf("Invalid file number header\n");
      return(-1);
    }

    /* Show disk information */
    memcpy(S,buf+16,4);S[4]='\0';
    Files=buf[57];
    printf
    (
      "DISK '%-4s'  Side %c  Files %d  Maker $%02X  Version $%02X\n",
      S,(buf[21]&1)+'A',Files,buf[15],buf[20]
    );

    /* Scan through the files */
    for(I=0;(I<Files)&&(fread(buf,1,16,fh)==16);I++)
    {
      /* Check if header block ID is valid */
      if(buf[0]!=3)
      {
        printf("Invalid file header $%02X\n",buf[0]);
        return(-1);
      }

      /* Get name, data location, and size */
      strncpy(S,buf+3,8);S[8]='\0';
      Start=buf[11]+256*buf[12];
      Size=buf[13]+256*buf[14];

      /* Check if data block ID is valid */
      J=fgetc(fh);
      if(J!=4)
      {
        printf("Invalid file header $%02X\n",J);
        return(-1);
      }

      /* List the file */
      sprintf(T,"$%02X?",buf[15]);
      printf
      (
        "%03d $%02X '%-8s' $%04X-$%04X [%s]\n",
        buf[1],buf[2],S,Start,Start+Size-1,
        buf[15]>2? T:
        buf[15]>1? "PICTURE":
        buf[15]>0? "TILES":
                   "CODE"
      );

      /* Seek over the data */
      fseek(fh,Size,SEEK_CUR);
    }

    /* Done with a disk */
  }

  /* Done */

	return(0);
}

