/*
psx.c - Playstation (One) support for uCON64

written by 2001 NoisyB (noisyb@gmx.net)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "psx.h"

#include "../patch/ppf.h"
#include "../patch/xps.h"
#include "../patch/pal4u.h"

#include "../backup/cdrw.h"

#define psx_ISO_HEADER_START 0x8008
#define psx_RAW_HEADER_START 0x9318
#define psx_HEADER_LEN 2048


int psx_init(struct ucon64_ *rom)
{
//TODO autodetect


#define ISODCL(from, to) (to - from + 1)

struct iso_primary_descriptor {
	char type			[ISODCL (  1,   1)]; /* 711 */
	char id				[ISODCL (  2,   6)];
	char version			[ISODCL (  7,   7)]; /* 711 */
	char unused1			[ISODCL (  8,   8)];
	char system_id			[ISODCL (  9,  40)]; /* achars */
	char volume_id			[ISODCL ( 41,  72)]; /* dchars */
	char unused2			[ISODCL ( 73,  80)];
	char volume_space_size		[ISODCL ( 81,  88)]; /* 733 */
	char unused3			[ISODCL ( 89, 120)];
	char volume_set_size		[ISODCL (121, 124)]; /* 723 */
	char volume_sequence_number	[ISODCL (125, 128)]; /* 723 */
	char logical_block_size		[ISODCL (129, 132)]; /* 723 */
	char path_table_size		[ISODCL (133, 140)]; /* 733 */
	char type_l_path_table		[ISODCL (141, 144)]; /* 731 */
	char opt_type_l_path_table	[ISODCL (145, 148)]; /* 731 */
	char type_m_path_table		[ISODCL (149, 152)]; /* 732 */
	char opt_type_m_path_table	[ISODCL (153, 156)]; /* 732 */
	char root_directory_record	[ISODCL (157, 190)]; /* 9.1 */
	char volume_set_id		[ISODCL (191, 318)]; /* dchars */
	char publisher_id		[ISODCL (319, 446)]; /* achars */
	char preparer_id		[ISODCL (447, 574)]; /* achars */
	char application_id		[ISODCL (575, 702)]; /* achars */
	char copyright_file_id		[ISODCL (703, 739)]; /* 7.5 dchars */
	char abstract_file_id		[ISODCL (740, 776)]; /* 7.5 dchars */
	char bibliographic_file_id	[ISODCL (777, 813)]; /* 7.5 dchars */
	char creation_date		[ISODCL (814, 830)]; /* 8.4.26.1 */
	char modification_date		[ISODCL (831, 847)]; /* 8.4.26.1 */
	char expiration_date		[ISODCL (848, 864)]; /* 8.4.26.1 */
	char effective_date		[ISODCL (865, 881)]; /* 8.4.26.1 */
	char file_structure_version	[ISODCL (882, 882)]; /* 711 */
	char unused4			[ISODCL (883, 883)];
	char application_data		[ISODCL (884, 1395)];
	char unused5			[ISODCL (1396, 2048)];
}toc;
  rom->console=ucon64_PSX;

  if(!argcmp(rom->argc,rom->argv,"-psx"))
  {
    rom->console=ucon64_UNKNOWN;
    return(-1);
  }

  strcpy(rom->title,psx_TITLE);

  rom->header_start = psx_RAW_HEADER_START;

  if(argcmp(rom->argc,rom->argv,"-iso"))rom->header_start = psx_ISO_HEADER_START;
  if(argcmp(rom->argc,rom->argv,"-raw"))rom->header_start = psx_RAW_HEADER_START;
  rom->header_len = psx_HEADER_LEN;

  quickfread(rom->header,rom->header_start,rom->header_len,rom->rom);
  quickfread(&toc,rom->header_start,sizeof(struct iso_primary_descriptor),rom->rom);

  sprintf(rom->copier,"%s",cdrw_TITLE);

  strcpy(rom->name,toc.volume_id);
  strcpy(rom->manufacturer,toc.publisher_id);
//  strcpy(rom->country,
//  strcpy(rom->misc,"");
//  strcpy(rom->misc,
    return(0);
}


int psx_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -psx		force recognition; NEEDED\n"
//	"  -iso		force image is ISO9660\n"
//	"  -raw		force image is MODE2_RAW/BIN\n"
	"  *		show info (default); ONLY $ROM=RAW_IMAGE\n"
//	"  -r2i		convert MODE2_RAW/BIN to ISO9660; $ROM=RAW_IMAGE\n"//net needed for psx
,psx_TITLE
);

pal4u_usage( argc, argv );
ppf_usage( argc, argv );
xps_usage( argc, argv );

#ifdef CD
  cdrw_raw_usage(argc,argv);
#endif
        
printf("\n");

return(0);
}

#ifdef CD
  int psx_rawt(struct ucon64_ *rom)
  {
    return( cdrw_raw_mktoc(rom) );
  }

  int psx_xraw(struct ucon64_ *rom)
  {
    return( (!access(rom->rom, F_OK)) ? cdrw_raw_write(rom) : cdrw_raw_read(rom) );
  }
#endif

/*
CD_ROM_XA


// Track 1
TRACK MODE2_RAW
NO COPY
DATAFILE "/root/src/ucon64/src/hentai.img" 00:06:12 // length in bytes: 1086624


// Track 2
TRACK MODE2_RAW
NO COPY
DATAFILE "/root/src/ucon64/src/hentai.img" #1086624 12:33:31 // length in bytes: 132902112


// Track 3
TRACK MODE2_RAW
NO COPY
DATAFILE "/root/src/ucon64/src/hentai.img" #133988736 12:52:71 // length in bytes: 136347792


// Track 4
TRACK MODE2_RAW
NO COPY
DATAFILE "/root/src/ucon64/src/hentai.img" #270336528 13:53:02 // length in bytes: 146945904


// Track 5
TRACK MODE2_RAW
NO COPY
DATAFILE "/root/src/ucon64/src/hentai.img" #417282432 27:25:30 // length in bytes: 290248560



*/
