/*
sms.c - Sega Master System/GameGear support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "sms.h"

#include "../backup/mgd.h"
#include "../backup/smd.h"

#define sms_ROM_HEADER_START 0x0
#define sms_ROM_HEADER_LEN 64

char buf[32768];
char buf2[32768];
char header[512];
//FILE *fh;
long x,y;

int sms_gge(struct ucon64_ *rom)
{
	if(!gameGenieEncodeGameGear(rom->rom, buf))
	{
		printf("%-12s = %s\n", rom->rom, buf);
	}
	else
	{
		printf("ERROR: %-12s is badly formed\n", rom->rom);
	}
	return(0);
}

int sms_ggd(struct ucon64_ *rom)
{
	if(!gameGenieDecodeGameGear(rom->rom, buf))
	{
		printf("%-12s = %s\n", rom->rom, buf);
	}
	else
	{
		printf("ERROR: %-12s is badly formed\n", rom->rom);
	}
	return(0);
}

int sms_gg(struct ucon64_ *rom)
{
	if(gameGenieDecodeGameGear(rom->file, buf)!=0)
	{
		printf("ERROR: %-12s is badly formed\n", rom->file);
		return(-1);
	}
	printf("%-12s = %s\n", rom->file, buf);

	sscanf(buf,"%lx:%lx:*",&x,&y);

	printf("\n");
	filehexdump(rom->rom,x+rom->buheader_len,1);

	buf[0]=(char)y;
	quickfwrite(buf,x+rom->buheader_len,1,filebackup(rom->rom),"r+b");

	filehexdump(rom->rom,x+rom->buheader_len,1);
	printf("\n");
	return(0);
}

int sms_mgd(struct ucon64_ *rom)
{
/*
The MGD2 only accepts certain filenames, and these filenames
must be specified in an index file, "MULTI-GD", otherwise the
MGD2 will not recognize the file.  In the case of multiple games
being stored in a single disk, simply enter its corresponding
MULTI-GD index into the "MULTI-GD" file.

Super Famiciom:

game size       # of files      names           MULTI-GD
================================================================
4M              1               SF4XXX.048      SF4XXX
4M              2               SF4XXXxA.078    SF4XXXxA
                                SF4XXXxB.078    SF4XXXxB
8M              1               SF8XXX.058      SF8XXX
                2               SF8XXXxA.078    SF8XXXxA
                                SF8XXXxB.078    SF8xxxxB
16M             2               SF16XXXA.078    SF16XXXA
                                SF16XXXB.078    SF16XXXB
24M             3               SF24XXXA.078    SF24XXXA
                                SF24XXXB.078    SF24XXXB
                                SF24XXXC.078    SF24XXXC

Mega Drive:

game size       # of files      names           MUTLI-GD
================================================================
1M              1               MD1XXX.000      MD1XXX
2M              1               MD2XXX.000      MD2XXX
4M              1               MD4XXX.000      MD4XXX
8M              1               MD8XXX.008      MD8XXX
16M             2               MD16XXXA.018    MD16XXXA
                                MD16XXXB.018    MD16XXXB
24M             3               MD24XXXA.038    MD24XXXA
                                MD24XXXB.038    MD24XXXB
                                MD24XXXC.038    MD24XXXC
32M             4               MD32XXXA.038    MD32XXXA
                                MD32XXXB.038    MD32XXXB
                                MD32XXXC.038    MD32XXXC
                                MD32XXXD.038    MD32XXXD

PC Engine:
game size       # of files      names           MUTLI-GD
================================================================
1M              1               PC1XXX.040      PC1XXX
2M              1               PC2XXX.040      PC2XXX
4M              1               PC4XXX.048      PC4XXX
8M              1               PC8XXX.058      PC8XXX


The Game Doctor does not use a 512 byte header like the SWC,
instead it uses specially designed filenames to distinguish
between multi files. I'm not sure if it used the filename for
information about the size of the image though.
<p>
Usually, the filename is in the format of: SFXXYYYZ.078
<p>
Where SF means Super Famicon, XX refers to the size of the
image in Mbit. If the size is only one character (i.e. 2, 4 or
8 Mbit) then no leading "0" is inserted.
<p>
YYY refers to a catalogue number in Hong Kong shops
identifying the game title. (0 is Super Mario World, 1 is F-
Zero, etc). I was told that the Game Doctor copier produces a
random number when backing up games.
<p>
Z indicates a multi file. Like XX, if it isn't used it's
ignored.
<p>
A would indicate the first file, B the second, etc. I am told
078 is not needed, but is placed on the end of the filename by
systems in Asia.
<p>
e.g. The first 16Mbit file of Donkey Kong Country (assuming it
  is cat. no. 475) would look like:  SF16475A.078
*/
	if(!rom->buheader_len)
        {
		printf("ERROR: Already in MGD format (use -hd to override)\n");
		return(-1);
	}
                                                
	strcpy(buf,findlwr(rom->rom)?"gg":"GG");
	strcat(buf,filenameonly(rom->rom));
	buf[findlast(buf,".")]=0;
	strcat(buf,"________");
	buf[7]='_';
	buf[8]=0;
	sprintf(buf2,"%s.%03lu",buf,(unsigned long)((quickftell(rom->rom)-rom->buheader_len)/MBIT));

	filecopy(rom->rom,rom->buheader_len,quickftell(rom->rom),filebackup(buf2),"wb");

	printf("Wrote output to %s\n", buf);
	return(0);
}

int sms_smd(struct ucon64_ *rom)
{
	if(rom->buheader_len!=0)
	{
		printf("ERROR: Already in SMD format (use -nhd to override)\n");
		return(-1);
	}

	memset(header,0,sizeof(header));
	header[1]=(((quickftell(rom->rom)-rom->buheader_len)/8192)/256);
	header[0]=(((quickftell(rom->rom)-rom->buheader_len)/8192)-(header[1]*256));
	header[8]=0xaa;
	header[9]=0xbb;
	header[10]=0x06;

	strcpy(buf,rom->rom);
	newext(buf,".SMD");

	quickfwrite(header,0,sizeof(header),filebackup(buf),"wb");

	filecopy(rom->rom,0,quickftell(rom->rom),buf,"ab");

        printf("Wrote output to %s\n", buf);
	return(0);
}

int sms_smds(struct ucon64_ *rom)
{
  memset(rom->buheader, 0, smd_HEADER_LEN);

  rom->buheader[0x00] = 0x00;
  rom->buheader[0x01] = 0x00; /* SRAM file */
  rom->buheader[0x02] = 0x00;
  rom->buheader[0x08] = 0xAA;
  rom->buheader[0x09] = 0xBB;
  rom->buheader[0x0A] = 0x07; /* SRAM file */

  strcpy(buf, rom->rom);
  newext(buf, ".TMP");
  rename(filebackup(rom->rom), buf);

  quickfwrite(rom->buheader, 0, smd_HEADER_LEN, rom->rom, "wb");
  filecopy(buf, 0, rom->bytes, rom->rom, "ab");
  printf("Wrote output to %s\n", rom->rom);

  remove(buf);
  return 0;
}

int sms_init(struct ucon64_ *rom)
{
rom->console=ucon64_SMS;

rom->buheader_len=0;
ucon64_dbsearch(rom);

if(rom->db_crc32==0)
{
/*	rom->buheader_len=smd_HEADER_LEN;
	rom->current_crc32=fileCRC32(rom->rom,rom->buheader_len);

	ucon64_dbsearch(rom);

	if(rom->db_crc32==0)
*/	{
		if(!argcmp(rom->argc,rom->argv,"-sms"))
		{
                        rom->console=ucon64_UNKNOWN;
                        return(-1);
		}
	}
}

  if(argcmp(rom->argc,rom->argv,"-hd"))rom->buheader_len=smd_HEADER_LEN;
  if(argcmp(rom->argc,rom->argv,"-nhd"))rom->buheader_len=0;

  strcpy(rom->title,sms_TITLE);
  strcpy(rom->copier,(!rom->buheader_len) ? mgd_TITLE : smd_TITLE);

  rom->header_start=sms_ROM_HEADER_START;
  rom->header_len=sms_ROM_HEADER_LEN;
  quickfread(rom->header,rom->header_start,rom->header_len,rom->rom);
  
  return(0);
}


int
gameGenieDecodeGameGear(const char *in, char *out) {
  int address = 0;
  int value = 0;
  int check = 0;
  int haveCheck = 0;
  int i;

  if (strlen(in) == 11 && in[3] == '-' && in[7] == '-') {
    haveCheck = 1;
  } else if (strlen(in) == 7 && in[3] == '-') {
    haveCheck = 0;
  } else {
    return -1;
  }
  for (i=0; i<strlen(in); ++i)
    if (in[i] != '-' && !isxdigit((int) in[i]))
      return -1;
  if (hexValue(in[6]) < 8) return -1;

  value = hexByteValue(in[0], in[1]);

  address = (hexValue(in[2]) << 8) | (hexValue(in[4]) << 4) | hexValue(in[5]) |
    ((~hexValue(in[6]) & 0xf) << 12);
  
  if (haveCheck) {
    check = hexByteValue(in[8], in[10]);
    check = ~check;
    check = (((check & 0xfc) >> 2) | ((check & 0x03) << 6)) ^ 0x45;
    sprintf(out, "%04X:%02X:%02X", address, value, check);
  } else {
    sprintf(out, "%04X:%02X", address, value);
  }

  return 0;
}

int
gameGenieEncodeGameGear(const char *in, char *out) {
  int check = 0;
  int haveCheck = 0;
  int i;

  if (strlen(in) == 10 && in[4] == ':' && in[7] == ':') {
    haveCheck = 1;
  } else if (strlen(in) == 7 && in[4] == ':') {
    haveCheck = 0;
  } else {
    return -1;
  }
  for (i=0; i<strlen(in); ++i)
    if (in[i] != ':' && !isxdigit((int) in[i]))
      return -1;
  if (hexValue(in[0]) > 7) return -1;

  out[0] = toupper(in[5]);
  out[1] = toupper(in[6]);
  out[2] = toupper(in[1]);
  out[3] = '-';
  out[4] = toupper(in[2]);
  out[5] = toupper(in[3]);
  out[6] = hexDigit(~hexValue(in[0]) & 0xf);
  out[7] = 0;

  if (haveCheck) {
    check = hexByteValue(in[8], in[10]);
    check = ~check;
    check = (((check & 0xfc) >> 2) | ((check & 0x03) << 6)) ^ 0x45;

    check = hexByteValue(in[8], in[9]);
    check ^= 0x45;
    check = ~(((check & 0xc0) >> 6) | ((check & 0x3f) << 2));
    i = (check & 0xf0) >> 4;
    out[7] = '-';
    out[8] = hexDigit(i & 0xf);
    out[9] = hexDigit((i ^ 8) & 0xf);
    out[10] = hexDigit(check & 0xf);
    out[11] = 0;
  }

  return 0;
}




int sms_usage(int argc,char *argv[])
{
printf("%s\n"
	"  -sms		force recognition"
#ifndef DB
"; NEEDED"
#endif
	"\n  -hd		force ROM has header (+512 Bytes)\n"
	"  -nhd		force ROM has no header (MGD2/MGH/RAW)\n"
	"  *		show info (default)\n"
//	"TODO: show more detailed ROM info\n"
	"  -mgd		convert to Multi Game*/MGD2/MGH/RAW\n"
	"  -smd		convert to Super Magic Drive/SMD (+512 Bytes)\n"
	"  -smds         convert Emulator (*.srm) SRAM to Super Magic Drive/SMD\n"
//	"TODO:  -chk	fix ROM checksum\n"
	"  -gge		encode GameGenie code; $ROM=AAAA:VV or $ROM=AAAA:VV:CC\n"
	"  -ggd		decode GameGenie code; $ROM=XXX-XXX or $ROM=XXX-XXX-XXX\n"
	"  -gg		apply GameGenie code (permanent); $FILE=CODE-CODE\n"
,sms_TITLE
);

#ifdef BACKUP
  smd_usage(argc,argv);
#endif

printf("\n");

return(0);
}
