/*
snes.c - Super Nintendo support for uCON64

written by 1999 - 2001 NoisyB (noisyb@gmx.net)
                  2001 dbjh


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "snes.h"

#include "../patch/bsl.h"
#include "../patch/ips.h"

#include "../backup/smc.h"
#include "../backup/mgd.h"
#include "../backup/swc.h"
#include "../backup/fig.h"

#define snes_ROM_HEADER_START 32688
#define snes_ROM_HEADER_LEN 48
#define snes_HIROM 32768
#define snes_NAME_LEN 21
#define ALT_HILO                                // use Snes9x' Hi/LoROM detection method

int snes_chksum(struct ucon64_ *rom, unsigned char *rom_buffer);
int snes_deinterleave(struct ucon64_ *rom, unsigned char *rom_buffer);
int gameGenieEncodeSNES(const char *in, char *out);
int gameGenieDecodeSNES(const char *in, char *out);
#ifdef  ALT_HILO
int scoreLoROM(unsigned char *rom_buffer, int rom_size);
int scoreHiROM(unsigned char *rom_buffer, int rom_size);
#endif

int snes_hirom = 0, snes_splitted = 0, snes_sramsize = 0, force_interleaved = 0,
    rom_is_top = 0;                             // flag for interleaved ROM dump
                                                //  of "Tales of Phantasia"
int snes_dint(struct ucon64_ *rom)
{
  FILE *srcfile, *destfile;
  unsigned char *buffer;
  char backup_name[FILENAME_MAX];
  int size = rom->bytes - rom->buheader_len;

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  puts("Converting to deinterleaved format...");

  if (!(buffer = (unsigned char *) malloc(size)))
  {
    fprintf(STDERR, "Not enough memory for ROM buffer (%d bytes)\n", size);
    exit(1);
  }
  strcpy(backup_name, rom->rom);
  newext(backup_name, ".BAK");
  rename(rom->rom, backup_name);
  if ((srcfile = fopen(backup_name, "rb")) == NULL)
  {
    printf("ERROR: could not open %s\n", backup_name);
    return -1;
  }
  if ((destfile = fopen(rom->rom, "wb")) == NULL)
  {
    printf("ERROR: could not open %s\n", rom->rom);
    return -1;
  }

  fseek(srcfile, rom->buheader_len, SEEK_SET);
  if (!fread(buffer, size, 1, srcfile))
  {
    free(buffer);
    fclose(srcfile);
    fclose(destfile);
    return -1;
  }

  snes_deinterleave(rom, buffer);

  if (rom->buheader_len)
  {
    if (rom->buheader[8] == 0xaa && rom->buheader[9] == 0xbb && rom->buheader[10] == 4)
    {                                           // SWC header
      rom->buheader[2] &= ~0x30;
      rom->buheader[2] |= snes_hirom ? 0x30 : 0x00;
    }
    else                                        // FIG header
      rom->buheader[3] = snes_hirom ? 0x80 : 0x00;
    fwrite(rom->buheader, 1, smc_HEADER_LEN, destfile);
  }
  fwrite(buffer, size, 1, destfile);

  free(buffer);
  fclose(srcfile);
  fclose(destfile);

  puts("Done");
  return 0;
}

int snes_gge(struct ucon64_ *rom)
{
	char buf[FILENAME_MAX];

        if(!gameGenieEncodeSNES(rom->rom, buf))
        {
                printf("%-12s = %s\n", rom->rom, buf);
        }
        else
        {
                printf("ERROR: %-12s is badly formed\n", rom->rom);
        }
        return(0);
}

int snes_ggd(struct ucon64_ *rom)
{
	char buf[FILENAME_MAX];

        if(!gameGenieDecodeSNES(rom->rom, buf))
        {
                printf("%-12s = %s\n", rom->rom, buf);
        }
        else
        {
                printf("ERROR: %-12s is badly formed\n", rom->rom);
        }
        return(0);
}


int snes_col(struct ucon64_ *rom)
{
/*
The Nintendo Super Famicom is capable of displaying 256 colours from a
palette of 32,768. These 256 colours are split into 8 palettes of 32 colours
each.

To change the colours the following needs to be done:

Loading the palette control register ($2121) with the colour number you wish
to change (0-255, 0=background).
Then load the colour into the palette data register first the low 8 bits,
followed by the high 7 bits (this gives you the maximum 32768 colours
possible $0000-$7fff).

Colour data is made up of 3 components (Red,Green,Blue) each of 5 bits (The
Amiga uses exactly the same system, but only using 4 bits per component).
Saying that, Nintendo being the stupid japanese idiots they are decided that
R,G,B wasn't alphabetically correct and so opted to store the bits as B,G,R.

                      00000 00000 00000
                      \   / \   / \   /
                       \ /   \ /   \ /
                        B     G     R

Examples:
~~~~~~~~~
         11111 00000 00000 = $7C00 (Bright Blue)
         00000 11111 00000 = $03E0 (Bright Green)
         00000 00000 11111 = $001F (Bright Red)
         00000 00000 00000 = $0000 (Black)
         11111 11111 11111 = $7FFF (White)

Easy, isn't it?? (But remember to load the lowest 8 bits first, then the top
                  7 bits).
*/


int r,g,b;
unsigned int col;

        sscanf(rom->rom,"%x",&col);

        r=(col&0xff0000)/0x10000;
        g=(col&0xff00)/0x100;
        b=col&0xff;

        printf("0x%02x%02x%02x (html) == ",r,g,b);

        r=(r*0x1f)/0xff;
        g=(g*0x1f)/0xff;
        b=(b*0x1f)/0xff;

        col=b;
        col=col<<5;
        col=col+g;
        col=col<<5;
        col=col+r;

        printf("0x%02x%02x (snes)\n",col&0xff,(col&0x7f00)/0x100);

        sscanf(rom->rom,"%x",&col);

        if((unsigned long)col<0xff7f+1)
        {
                printf("0x%04x (snes) == ",col&0xff7f);

                col=((col&0x7f)*0x100)+((col&0xff00)/0x100);

                r=col&0x1f;
                g=(col&(0x1f<<5))>>5;
                b=(col&(0x1f<<10))>>10;

                r=r*0xff/0x1f;
                g=g*0xff/0x1f;
                b=b*0xff/0x1f;

                printf("0x%02x%02x%02x (html)\n",r,g,b);
        }
        printf("\n");
        return(0);
}

int snes_swcs(struct ucon64_ *rom)
{
  char buf[FILENAME_MAX];

  memset(rom->buheader, 0, smc_HEADER_LEN);

  rom->buheader[8] = 0xaa;
  rom->buheader[9] = 0xbb;
  rom->buheader[10] = 0x05;

  strcpy(buf, rom->rom);
  newext(buf, ".TMP");
  rename(filebackup(rom->rom), buf);

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, rom->rom, "wb");
  filecopy(buf, 0, rom->bytes, rom->rom, "ab");
  printf("Wrote output to %s\n", rom->rom);

  remove(buf);
  return 0;
}

int snes_figs(struct ucon64_ *rom)
{
  char buf[FILENAME_MAX];

  memset(rom->buheader, 0, smc_HEADER_LEN);

  rom->buheader[0] = 0x04;

  strcpy(buf, rom->rom);
  newext(buf, ".TMP");
  rename(filebackup(rom->rom), buf);

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, rom->rom, "wb");
  filecopy(buf, 0, rom->bytes, rom->rom, "ab");
  printf("Wrote output to %s\n", rom->rom);

  remove(buf);
  return 0;
}

int snes_ufos(struct ucon64_ *rom)
{
  char buf[FILENAME_MAX];

  memset(rom->buheader, 0, smc_HEADER_LEN);

  rom->buheader[8]  = 'S';
  rom->buheader[9]  = 'U';
  rom->buheader[10] = 'P';
  rom->buheader[11] = 'E';
  rom->buheader[12] = 'R';
  rom->buheader[13] = 'U';
  rom->buheader[14] = 'F';
  rom->buheader[15] = 'O';

  strcpy(buf, rom->rom);
  newext(buf, ".TMP");
  rename(filebackup(rom->rom), buf);

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, rom->rom, "wb");
  filecopy(buf, 0, rom->bytes, rom->rom, "ab");
  printf("Wrote output to %s\n", rom->rom);

  remove(buf);
  return 0;
}

int snes_gg(struct ucon64_ *rom)
{
	long x,y;
	char buf[4096];

        if(snes_splitted!=0)
        {
                printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
                return(-1);
        }

        if(gameGenieDecodeSNES(rom->file, buf)!=0)
        {
                printf("ERROR: %-12s is badly formed\n", rom->file);
                return(-1);
        }
        printf("%-12s = %s\n", rom->file, buf);

        sscanf(buf,"%lx:%lx:*",&x,&y);

        printf("\n");
        filehexdump(rom->rom,x+rom->buheader_len,1);

        buf[0]=(char)y;
        quickfwrite(buf,x+rom->buheader_len,1,filebackup(rom->rom),"r+b");

        filehexdump(rom->rom,x+rom->buheader_len,1);
        printf("\n");
        return(0);
}



int snes_smc(struct ucon64_ *rom)
{
  int size;
  char src_name[FILENAME_MAX], buf[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  if (snes_hirom != 0)
    printf("NOTE: this game might not work with a Super Magicom because it's HiROM\n");

  memset(rom->buheader, 0, smc_HEADER_LEN);
  size = rom->bytes - rom->buheader_len;
  rom->buheader[0] = size/8192;
  rom->buheader[1] = size/8192 >> 8;
  rom->buheader[8] = 0xaa;
  rom->buheader[9] = 0xbb;
  rom->buheader[10] = 0x04;

  strcpy(buf, rom->rom);
  newext(buf, ".SMC");
  strcpy(src_name, rom->rom);
  if (!access(buf, F_OK))
  {                                             // ROM name with smc extension already exists
    filebackup(buf);
    if (!extcmp(src_name, "smc"))
    {
      strcat(src_name, (findlwr(src_name) ? ".bak" : ".BAK" ));
      printf("Wrote backup to %s\n", src_name);
    }
    else
      printf("Wrote backup to %s%s\n", buf, (findlwr(buf) ? ".bak" : ".BAK" ));
  }                                             // must match with what filebackup() does

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, buf, "wb");
  filecopy(src_name, rom->buheader_len, size, buf, "ab");
  printf("Wrote output to %s\n", buf);

  return 0;
}


int snes_fig(struct ucon64_ *rom)
{
/*
Offset Function
  0    Lower 8 bits of size word
  1    Upper 8 bits of size word
  2    40h - Multi image
       00h - Last image in set (or single image)
  3    80h - if HiROM
       00h - if LoROM
  4    If using DSP1 microchip:
           FDh - If using SRAM (SRAM size>0)
           47h - If no SRAM (SRAM size=0)
       77h - If not using DSP1 and no SRAM (SRAM size=0)
  5    If using DSP1 microchip:
           82h - If using SRAM (SRAM size>0)
           83h - If no SRAM (SRAM size=0)
       83h - If not using DSP1 and no SRAM (SRAM size=0)
*/
  int size, x;
  char src_name[FILENAME_MAX], buf[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  memset(rom->buheader, 0, smc_HEADER_LEN);
  size = rom->bytes - rom->buheader_len;
  rom->buheader[0] = size/8192;
  rom->buheader[1] = size/8192 >> 8;
  rom->buheader[2] = snes_splitted ? 0x40 : 0x00;
  rom->buheader[3] = snes_hirom ? 0x80 : 0x00;

  x = quickfgetc(rom->rom, rom->header_start + rom->buheader_len + 38) & 0x07;

  rom->buheader[4] = (x == 3 || x == 4 || x == 5 ) ?
                       (snes_sramsize ? 0xfd : 0x47) : // DSP1
                       (snes_sramsize ? 0x00 : 0x77);  // no DSP1
  rom->buheader[5] = (x == 3 || x == 4 || x == 5 ) ?
                       (snes_sramsize ? 0x82 : 0x83) : // DSP1
                       (snes_sramsize ? 0x00 : 0x83);  // no DSP1

  strcpy(buf, rom->rom);
  newext(buf, ".FIG");
  strcpy(src_name, rom->rom);
  if (!access(buf, F_OK))
  {                                             // ROM name with fig extension already exists
    filebackup(buf);
    if (!extcmp(src_name, "fig"))
    {
      strcat(src_name, (findlwr(src_name) ? ".bak" : ".BAK" ));
      printf("Wrote backup to %s\n", src_name);
    }
    else
      printf("Wrote backup to %s%s\n", buf, (findlwr(buf) ? ".bak" : ".BAK" ));
  }                                             // must match with what filebackup() does

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, buf, "wb");
  filecopy(src_name, rom->buheader_len, size, buf, "ab");
  printf("Wrote output to %s\n", buf);

  return 0;
}


int snes_swc(struct ucon64_ *rom)
{
/*
    0      - Low byte of 8kB page count
    1      - High byte of 8kB page count
    2      - Emulation Mode Select
             Bit 7 6 5 4 3 2 1 0
                 x               : 0 = Run program in mode 1 (JMP RESET Vector)
                                 : 1 = Run in mode 0 (JMP $8000)
                   x             : 0 = Last file of the ROM dump (multi-file)
                                 : 1 = Multi-file (there is another file to follow)
                     x           : 0 = SRAM memory mapping mode 20, 1 = mode 21 (HiROM)
                       x         : 0 = DRAM memory mapping mode 20, 1 = mode 21 (HiROM)
                         x x     : 00 = 256kb SRAM, 01 = 64kb, 10 = 16kb, 11 = no SRAM
                             x   : 0 = Disable, 1 = Enable external cartridge
                               x : reserved
    3-7    - 00, reserved
    8      - File ID code 1 (0xAA)
    9      - File ID code 2 (0xBB)
    10     - File type; check this byte only if ID 1 & 2 match
             02 : Magic Griffin program (PC Engine)
             03 : Magic Griffin SRAM data
             04 : SNES program
             05 : SWC & SMC password, SRAM data
             06 : Megadrive program
             07 : SMD SRAM data
    11-511 - 00, reserved
*/
  int size;
  char src_name[FILENAME_MAX], buf[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  memset(rom->buheader, 0, smc_HEADER_LEN);
  size = rom->bytes - rom->buheader_len;
  rom->buheader[0] = size/8192;
  rom->buheader[1] = size/8192 >> 8;

  rom->buheader[2] = snes_splitted ? 0x40 : 0x00;
  rom->buheader[2] |= snes_hirom ? 0x30 : 0x00;
  switch (snes_sramsize)
  {
    // bit 3 & 2 already ok for 32*1024kB SRAM size
    case 8*1024:
      rom->buheader[2] |= 0x04;
      break;
    case 2*1024:
      rom->buheader[2] |= 0x08;
      break;
    case 0:
      rom->buheader[2] |= 0x0c;
      break;
  }

  rom->buheader[8] = 0xaa;
  rom->buheader[9] = 0xbb;
  rom->buheader[10] = 0x04;

  strcpy(buf, rom->rom);
  newext(buf, ".SWC");
  strcpy(src_name, rom->rom);
  if (!access(buf, F_OK))
  {                                             // ROM name with swc extension already exists
    filebackup(buf);
    if (!extcmp(src_name, "swc"))
    {
      strcat(src_name, (findlwr(src_name) ? ".bak" : ".BAK" ));
      printf("Wrote backup to %s\n", src_name);
    }
    else
      printf("Wrote backup to %s%s\n", buf, (findlwr(buf) ? ".bak" : ".BAK" ));
  }                                             // must match with what filebackup() does

  quickfwrite(rom->buheader, 0, smc_HEADER_LEN, buf, "wb");
  filecopy(src_name, rom->buheader_len, size, buf, "ab");
  printf("Wrote output to %s\n", buf);

  return 0;
}



int snes_mgd(struct ucon64_ *rom)
{
/*
The MGD2 only accepts certain filenames, and these filenames
must be specified in an index file, "MULTI-GD", otherwise the
MGD2 will not recognize the file.  In the case of multiple games
being stored in a single disk, simply enter its corresponding
MULTI-GD index into the "MULTI-GD" file.

Super Famiciom:

game size       # of files      names           MULTI-GD
================================================================
4M              1               SF4XXX.048      SF4XXX
4M              2               SF4XXXxA.078    SF4XXXxA
                                SF4XXXxB.078    SF4XXXxB
8M              1               SF8XXX.058      SF8XXX
                2               SF8XXXxA.078    SF8XXXxA
                                SF8XXXxB.078    SF8xxxxB
16M             2               SF16XXXA.078    SF16XXXA
                                SF16XXXB.078    SF16XXXB
24M             3               SF24XXXA.078    SF24XXXA
                                SF24XXXB.078    SF24XXXB
                                SF24XXXC.078    SF24XXXC

Mega Drive:

game size       # of files      names           MUTLI-GD
================================================================
1M              1               MD1XXX.000      MD1XXX
2M              1               MD2XXX.000      MD2XXX
4M              1               MD4XXX.000      MD4XXX
8M              1               MD8XXX.008      MD8XXX
16M             2               MD16XXXA.018    MD16XXXA
                                MD16XXXB.018    MD16XXXB
24M             3               MD24XXXA.038    MD24XXXA
                                MD24XXXB.038    MD24XXXB
                                MD24XXXC.038    MD24XXXC
32M             4               MD32XXXA.038    MD32XXXA
                                MD32XXXB.038    MD32XXXB
                                MD32XXXC.038    MD32XXXC
                                MD32XXXD.038    MD32XXXD

PC Engine:
game size       # of files      names           MUTLI-GD
================================================================
1M              1               PC1XXX.040      PC1XXX
2M              1               PC2XXX.040      PC2XXX
4M              1               PC4XXX.048      PC4XXX
8M              1               PC8XXX.058      PC8XXX


The Game Doctor does not use a 512 byte header like the SWC,
instead it uses specially designed filenames to distinguish
between multi files. I'm not sure if it used the filename for
information about the size of the image though.

Usually, the filename is in the format of: SFXXYYYZ.078

Where SF means Super Famicon, XX refers to the size of the
image in Mbit. If the size is only one character (i.e. 2, 4 or
8 Mbit) then no leading "0" is inserted.

YYY refers to a catalogue number in Hong Kong shops
identifying the game title. (0 is Super Mario World, 1 is F-
Zero, etc). I was told that the Game Doctor copier produces a
random number when backing up games.

Z indicates a multi file. Like XX, if it isn't used it's
ignored.

A would indicate the first file, B the second, etc. I am told
078 is not needed, but is placed on the end of the filename by
systems in Asia.

e.g. The first 16Mbit file of Donkey Kong Country (assuming it
  is cat. no. 475) would look like:  SF16475A.078
*/
        char mgh[32], buf[FILENAME_MAX], buf2[4096];
        int x;

        if(snes_splitted!=0)
        {
                printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
                return(-1);
        }
/*
        if(!rom->buheader_len)
        {
                printf("ERROR: Already in MGD format (use -hd to override)\n");
                return(-1);
        }
*/
        strcpy(buf,findlwr(rom->rom)?"sf":"SF");
        strcpy(buf2,rom->rom);
        strcat(buf,filenameonly(buf2));
        buf[findlast(buf,".")]=0;
        strcat(buf,"________");
        buf[7]='_';
        buf[8]=0;
        sprintf(buf2,"%s.%03lu", buf, (unsigned long) (rom->bytes - rom->buheader_len)/MBIT);

        filecopy(rom->rom, rom->buheader_len, rom->bytes, filebackup(buf2), "wb");
/*
        return(0);
}

int snes_mgh(struct ucon64_ *rom)
{
*/
        if(snes_splitted!=0)
        {
                printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
                return(-1);
        }
/*
        if(!rom->buheader_len)
        {
                printf("ERROR: Already in MGH format (use -hd to override)\n");
                return(-1);
        }
*/
/*      strcpy(buf,findlwr(rom->rom)?"sf":"SF");
        strcpy(buf2,rom->rom);
        strcat(buf,filenameonly(buf2));
        buf[findlast(buf,".")]=0;
        strcat(buf,"________");
        buf[7]='_';
        buf[8]=0;
        sprintf(buf2,"%s.%03lu" ,buf
                                ,(unsigned long)(rom->bytes-rom->buheader_len)/MBIT
                        );

        filecopy(       rom->rom
                        ,rom->buheader_len
                        ,rom->bytes
                        ,filebackup(buf2)
                        ,"wb"
        );
*/
//create MGH name file

        memset(mgh,0,sizeof(mgh));
        mgh[0]='M';
        mgh[1]='G';
        mgh[2]='H';
        mgh[3]=0x1a;
        mgh[31]=0xff;

        for(x=16;x<31;x++)
        {
                mgh[x]=quickfgetc(buf2, rom->header_start + rom->buheader_len + 16 + x);
        }

        strcpy(buf,buf2);
        newext(buf,".MGH");

        quickfwrite(&mgh,0,sizeof(mgh),buf,"wb");

        return(0);
}


int snes_gd3(struct ucon64_ *rom)
{
	char buf[FILENAME_MAX], buf2[4096];

        if(snes_splitted!=0)
        {
                printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
                return(-1);
        }
/*
        if(!rom->buheader_len)
        {
                printf("ERROR: Already in GD3 format (use -hd to override)\n");
                return(-1);
        }
*/
        sprintf(buf,"%s%02ld"   ,findlwr(rom->rom)?"sf":"SF"
                                ,(unsigned long)(rom->bytes-rom->buheader_len)/MBIT);
        strcat(buf,filenameonly(rom->rom));
        buf[findlast(buf,".")]=0;
        strcat(buf,"________");
        buf[7]='X';
        buf[8]=0;
        strcpy(buf2,buf);

        filecopy(rom->rom, rom->buheader_len, rom->bytes, filebackup(buf2), "wb");

        return(0);
}

int snes_gdf(struct ucon64_ *rom)
{
        if (!rom->buheader_len)
        {
           printf("ERROR: this ROM has no header and therefore it does not need to be fixed\n");
           return(-1);
        }

        rom->buheader[0x10]=0x81;
        rom->buheader[0x29]=0x0c;
        rom->buheader[0x30]=0x0c;

        quickfwrite(rom->buheader,0,smc_HEADER_LEN,filebackup(rom->rom),"r+b");

        return(0);
}


int snes_j(struct ucon64_ *rom)
{
  char buf[FILENAME_MAX], buf2[4096];
  int size = 0;

  strcpy(buf, rom->rom);
  strcpy(buf2, rom->rom);
  newext(buf2, ".TMP");

  filecopy(buf, 0, rom->buheader_len, filebackup(buf2), "wb"); // copy header (if any)
  while (filecopy(buf, rom->buheader_len, quickftell(buf), buf2, "ab") != -1)
  {
    size += quickftell(buf);
    buf[(!rom->buheader_len) ?
        (findlast(buf, ".") - 1) :              // without header (see code of "-s")
        (findlast(buf, ".") + 1)]++;            // with header (see code of "-s")
  }

  if (rom->buheader_len)
  {                                             // fix header
    quickfputc(buf2, 0, size/8192, "r+b");      // # 8K blocks low byte
    quickfputc(buf2, 1, size/8192 >> 8, "r+b"); // # 8K blocks high byte
    quickfputc(buf2, 2, quickfgetc(buf2, 2) & ~0x40, "r+b"); // last file -> clear bit 6
  }

  return 0;
}


int snes_s(struct ucon64_ *rom)
{
  char buf[FILENAME_MAX], buf2[4096];
  int n4Mbparts, surplus4Mb, x;

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  n4Mbparts = (rom->bytes - rom->buheader_len)/(4*MBIT);
  surplus4Mb = (rom->bytes - rom->buheader_len)%(4*MBIT);

  if (!rom->buheader_len)
  {
    strcpy(buf, findlwr(rom->rom) ? "sf" : "SF");
    strcat(buf, filenameonly(rom->rom));
    buf[findlast(buf, ".")] = 0;
    strcat(buf, "________");
    buf[7] = findlwr(buf) ? 'a' : 'A';
    buf[8] = 0;
    sprintf(buf2, "%s.%03lu", buf, (unsigned long) (rom->bytes - rom->buheader_len)/MBIT);

    for (x = 0; x < n4Mbparts; x++)
    {
      filecopy(rom->rom, x*4*MBIT, 4*MBIT, filebackup(buf2), "wb");
      buf2[findlast(buf2, ".") - 1]++;
    }

    if (surplus4Mb != 0)
      filecopy(rom->rom, x*4*MBIT, surplus4Mb, filebackup(buf2), "wb");

    return 0;
  }
  else
  {
    strcpy(buf, rom->rom);
    newext(buf, ".1");

    rom->buheader[0] = 4*MBIT/8192;
    rom->buheader[1] = 4*MBIT/8192 >> 8;
    // if header[2], bit 6 == 0 -> SWC knows this is the last file of the ROM
    rom->buheader[2] |= 0x40;
    for (x = 0; x < n4Mbparts; x++)
    {
      if (surplus4Mb == 0 && x == n4Mbparts - 1)
        rom->buheader[2] &= ~0x40;              // last file -> clear bit 6

      quickfwrite(rom->buheader, 0, smc_HEADER_LEN, filebackup(buf), "wb");
      filecopy(rom->rom, x*4*MBIT + rom->buheader_len, 4*MBIT, buf, "ab");

      buf[findlast(buf, ".") + 1]++;
    }

    if (surplus4Mb != 0)
    {
      rom->buheader[0] = surplus4Mb/8192;
      rom->buheader[1] = surplus4Mb/8192 >> 8;
      rom->buheader[2] &= ~0x40;                // last file -> clear bit 6

      quickfwrite(rom->buheader, 0, smc_HEADER_LEN, filebackup(buf), "wb");
      filecopy(rom->rom, x*4*MBIT + rom->buheader_len, surplus4Mb, buf, "ab");
    }

    return 0;
  }
}

int snes_p(struct ucon64_ *rom)
{
  char buf[MAXBUFSIZE];

  if (snes_splitted!=0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return(-1);
  }

  filepad(filebackup(buf), rom->buheader_len, MBIT);
  return 0;
}


int snes_k(struct ucon64_ *rom)
{
/*
Some SNES games check to see how much SRAM is connected to the
SNES as a form of copy protection. As most copiers have
256kbits standard, the game will know it's running on a backup
unit and stop to prevent people copying the games. However,
the newer copiers get around this detection somehow.

(original uCON)
        8F/9F xx xx 70 CF/DF xx xx 70 D0
becomes:8F/9F xx xx 70 CF/DF xx xx 70 EA EA     // if snes_sramsize == 64kbits
becomes:8F/9F xx xx 70 CF/DF xx xx 70 80

        8F/9F xx xx 30/31/32/33 CF/DF xx xx 30/31/32/33 D0
becomes:8F/9F xx xx 30/31/32/33 CF/DF xx xx 30/31/32/33 80

        8F/9F xx xx 30/31/32/33 CF/DF xx xx 30/31/32/33 F0
becomes:8F/9F xx xx 30/31/32/33 CF/DF xx xx 30/31/32/33 EA EA

        8F/9F xx xx 30/31/32/33 AF xx xx 30/31/32/33 C9 xx xx D0
becomes:8F/9F xx xx 30/31/32/33 AF xx xx 30/31/32/33 C9 xx xx 80

(uCON64)
                                                - Super Metroid
        a9 00 00 a2 fe 1f df 00 00 70 d0        // lda #$0000; ldx #$1ffe; cmp $700000,x; bne ...
becomes:a9 00 00 a2 fe 1f df 00 00 70 ea ea     // lda #$0000; ldx #$1ffe; cmp $700000,x; nop; nop
*/
  FILE *srcfile, *destfile;
  int bytesread;
  char buffer[32*1024], backup_name[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  puts("Attempting crack...");

  strcpy(backup_name, rom->rom);
  newext(backup_name, ".BAK");
  rename(rom->rom, backup_name);
  if ((srcfile = fopen(backup_name, "rb")) == NULL)
  {
    printf("ERROR: could not open %s\n", backup_name);
    return -1;
  }
  if ((destfile = fopen(rom->rom, "wb")) == NULL)
  {
    printf("ERROR: could not open %s\n", rom->rom);
    return -1;
  }
  if (rom->buheader_len)                        // copy header (if present)
  {
    fseek(srcfile, rom->buheader_len, SEEK_SET);
    fwrite(rom->buheader, 1, rom->buheader_len, destfile);
  }

  while ((bytesread = fread(buffer, 1, 32*1024, srcfile)))
  {                                             // '!' == ASCII 33 (\x21), '*' == 42 (\x2a)
    if (snes_sramsize == 8*1024)       // 8kB == 64kb
      change_string("!**\x70!**\x70\xd0", 9, '*', '!', "\xea\xea", 2, buffer, bytesread, 0,
                    "\x8f\x9f", 2, "\xcf\xdf", 2);
    else
      change_string("!**\x70!**\x70\xd0", 9, '*', '!', "\x80", 1, buffer, bytesread, 0,
                    "\x8f\x9f", 2, "\xcf\xdf", 2);

    change_string("!**!!**!\xd0", 9, '*', '!', "\x80", 1, buffer, bytesread, 0,
                  "\x8f\x9f", 2, "\x30\x31\x32\x33", 4, "\xcf\xdf", 2, "\x30\x31\x32\x33", 4);
    change_string("!**!!**!\xf0", 9, '*', '!', "\xea\xea", 2, buffer, bytesread, 0,
                  "\x8f\x9f", 2, "\x30\x31\x32\x33", 4, "\xcf\xdf", 2, "\x30\x31\x32\x33", 4);
    change_string("!**!\xaf**!\xc9**\xd0", 12, '*', '!', "\x80", 1, buffer, bytesread, 0,
                  "\x8f\x9f", 2, "\x30\x31\x32\x33", 4, "\x30\x31\x32\x33", 4);

    // uCON64
    change_string("\xa9\x00\x00\xa2\xfe\x1f\xdf\x00\x00\x70\xd0", 11, '*', '!', "\xea\xea", 2, buffer, bytesread, 0);

    fwrite(buffer, 1, bytesread, destfile);
  }
  fclose(srcfile);
  fclose(destfile);

  puts("Done");
  return 0;
}

int snes_f(struct ucon64_ *rom)
{
/*
The order is important. Don't touch this code if you don't know what you're doing!

Search for                      Replace with
(original uCON)
3F 21 29/89 10 C9 10 F0         3F 21 29/89 10 C9 10 80
3F 21 29/89 10 F0               3F 21 29/89 10 80
3F 21 00 29/89 10 F0            3F 21 00 29/89 10 80
3F 21 29/89 10 00 F0            3F 21 29/89 10 00 80
AD 3F 21 29 10 00 D0            AD 3F 21 29 10 00 80            - NTSC
AD 3F 21 89 10 00 D0            A9 10 00 89 10 00 D0            - NTSC Eric Cantona Football Challenge
AF 3F 21 00 29 10 D0            AF 3F 21 00 29 10 80            - NTSC
AF 3F 21 00 29 10 00 D0         AF 3F 21 00 29 10 00 EA EA      - NTSC
AD 3F 21 29 10 D0               AD 3F 21 29 10 EA EA            - PAL
AD 3F 21 29 10 F0               AD 3F 21 29 10 80               - PAL
AD 3F 21 89 10 D0               AD 3F 21 89 10 80               - PAL
AD 3F 21 29 10 C9 00 F0         AD 3F 21 29 10 C9 00 80         - PAL
AF 3F 21 00 29/89 10 00 F0      AF 3F 21 00 29/89 10 00 80      - PAL
AD 3F 21 29 10 CF BD FF 80 F0   AD 3F 21 29 10 CF BD FF 80 80   - PAL Konami Ganbare Goemon 2
AF 3F 21 EA 89 10 00 D0         A9 00 00 EA 89 10 00 D0         - PAL Super Famista 3
AD 3F 21 8D xx xx 29 10 8D      AD 3F 21 8D xx xx 29 00 8D      - PAL DragonBallZ 2
AF 3F 21 00 xx xx 29 10 00 D0   AF 3F 21 00 xx xx 29 10 00 EA EA- PAL Fatal Fury Special (Jap)
AD 3F 21 29 10 CF BD FF 00 F0   AD 3F 21 29 10 CF BD FF 00 80   - NTSC Tiny Toons Wacky Sports

(uCON64)
ad 3f 21 89 10 d0               ad 3f 21 89 10 ea ea
ad 3f 21 89 10 f0               ad 3f 21 89 10 80
ad 3f 21 29 10 c9 00 f0         ad 3f 21 29 10 c9 00 ea ea
ad 3f 21 29 10 c9 00 d0         ad 3f 21 29 10 c9 00 80
ad 3f 21 29 10 c9 10 f0         ad 3f 21 29 10 c9 10 80
ad 3f 21 29 10 c9 10 d0         ad 3f 21 29 10 c9 10 ea ea
*/
  FILE *srcfile, *destfile;
  int bytesread;
  char buffer[32*1024], backup_name[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  puts("Attempting NTSC/PAL fix...");

  strcpy(backup_name, rom->rom);
  newext(backup_name, ".BAK");
  rename(rom->rom, backup_name);
  if ((srcfile = fopen(backup_name, "rb")) == NULL)
  {
    printf("ERROR: could not open %s\n", backup_name);
    return -1;
  }
  if ((destfile = fopen(rom->rom, "wb")) == NULL)
  {
    printf("ERROR: could not open %s\n", rom->rom);
    return -1;
  }
  if (rom->buheader_len)                        // copy header (if present)
  {
    fseek(srcfile, rom->buheader_len, SEEK_SET);
    fwrite(rom->buheader, 1, rom->buheader_len, destfile);
  }

  while ((bytesread = fread(buffer, 1, 32*1024, srcfile)))
  {
    change_string("\x3F\x21\x02\x10\xC9\x10\xF0", 7, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0,
                  "\x29\x89", 2);
    change_string("\x3F\x21\x02\x10\xF0", 5, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0,
                  "\x29\x89", 2);
    change_string("\x3F\x21\x00\x02\x10\xF0", 6, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0,
                  "\x29\x89", 2);
    change_string("\x3F\x21\x02\x10\x00\xF0", 6, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0,
                  "\x29\x89", 2);
    change_string("\xAD\x3F\x21\x29\x10\x00\xD0", 7, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x89\x10\x00\xD0", 7, '\x01', '\x02', "\xA9\x10\x00", 3, buffer, bytesread, -6);
    change_string("\xAF\x3F\x21\x00\x29\x10\xD0", 7, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAF\x3F\x21\x00\x29\x10\x00\xD0", 8, '\x01', '\x02', "\xEA\xEA", 2, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x29\x10\xD0", 6, '\x01', '\x02', "\xEA\xEA", 2, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x29\x10\xF0", 6, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x89\x10\xD0", 6, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x29\x10\xC9\x00\xF0", 8, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAF\x3F\x21\x00\x02\x10\x00\xF0", 8, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0,
                  "\x29\x89", 2);
    change_string("\xAD\x3F\x21\x29\x10\xCF\xBD\xFF\x80\xF0", 10, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xAF\x3F\x21\xEA\x89\x10\x00\xD0", 8, '\x01', '\x02', "\xA9\x00\x00", 3, buffer, bytesread, -7);
    change_string("\xAD\x3F\x21\x8D\x01\x01\x29\x10\x8D", 9, '\x01', '\x02', "\x00", 1, buffer, bytesread, -1);
    change_string("\xAF\x3F\x21\x00\x01\x01\x29\x10\x00\xD0", 10, '\x01', '\x02', "\xEA\xEA", 2, buffer, bytesread, 0);
    change_string("\xAD\x3F\x21\x29\x10\xCF\xBD\xFF\x00\xF0", 10, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);

    // uCON64
    change_string("\xad\x3f\x21\x89\x10\xd0", 6, '\x01', '\x02', "\xea\xea", 2, buffer, bytesread, 0);
    change_string("\xad\x3f\x21\x89\x10\xf0", 6, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xad\x3f\x21\x29\x10\xc9\x00\xf0", 8, '\x01', '\x02', "\xea\xea", 2, buffer, bytesread, 0);
    change_string("\xad\x3f\x21\x29\x10\xc9\x00\xd0", 8, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xad\x3f\x21\x29\x10\xc9\x10\xf0", 8, '\x01', '\x02', "\x80", 1, buffer, bytesread, 0);
    change_string("\xad\x3f\x21\x29\x10\xc9\x10\xd0", 8, '\x01', '\x02', "\xea\xea", 2, buffer, bytesread, 0);

    fwrite(buffer, 1, bytesread, destfile);
  }
  fclose(srcfile);
  fclose(destfile);

  puts("Done");
  return 0;
}


int snes_l(struct ucon64_ *rom)
{
/*
The order is important. Don't touch this code if you don't know what you're doing!

Search for                      Replace with
(uCON64)
8c/8d/8e/8f 0d 42               9c 0d 42
01 0d 42                        00 0d 42
a9 01 85 0d                     a9 00 85 0d // special one (used by Konami and Jaleco? sometimes)
a2 01 86 0d                     a2 00 86 0d
a0 01 84 0d                     a0 00 84 0d

(original uCON)
A9/A2 01 8D/8E 0D 42            A9/A2 00 8D/8E 0D 42
A9 01 00 8D 0D 42               A9 00 00 8D 0D 42
A9 01 8F 0D 42 00               A9 00 8F 0D 42 00
*/
  int bytesread, x;
  char buffer[32*1024], backup_name[FILENAME_MAX];

  if (snes_splitted != 0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  puts("Attempting slowROM fix...");

  strcpy(backup_name, rom->rom);
  newext(backup_name, ".BAK");
  rename(rom->rom, backup_name);

  if (filecopy(backup_name, 0, rom->buheader_len, rom->rom, "wb") == -1)
  {
    printf("ERROR: could not open %s\n", backup_name);
    return -1;
  }

  x = 0;
  while ((bytesread = quickfread(buffer, x*(32*1024) + rom->buheader_len, 32*1024, backup_name)))
  {                                             // '!' == ASCII 33 (\x21), '*' == 42 (\x2a)
    change_string("!\x0d\x42", 3, '*', '!', "\x9c", 1, buffer, bytesread, -2,
                  "\x8c\x8d\x8e\x8f", 4);
    change_string("\x01\x0d\x42", 3, '*', '!', "\x00", 1, buffer, bytesread, -2);
    change_string("\xa9\x01\x85\x0d", 4, '*', '!', "\x00", 1, buffer, bytesread, -2);
    change_string("\xa2\x01\x86\x0d", 4, '*', '!', "\x00", 1, buffer, bytesread, -2);
    change_string("\xa0\x01\x84\x0d", 4, '*', '!', "\x00", 1, buffer, bytesread, -2);

    // original uCON
    change_string("!\x01!\x0D\x42", 5, '*', '!', "\x00", 1, buffer, bytesread, -3,
                  "\xA9\xA2", 2, "\x8D\x8E", 2);
    change_string("\xA9\x01\x00\x8D\x0D\x42", 6, '*', '!', "\x00", 1, buffer, bytesread, -4);
    change_string("\xA9\x01\x8F\x0D\x42\x00", 6, '*', '!', "\x00", 1, buffer, bytesread, -4);

    quickfwrite(buffer, x*(32*1024) + rom->buheader_len, bytesread, rom->rom, "ab");
    x++;
  }

  puts("Done");
  return 0;
}


int snes_n(struct ucon64_ *rom)
{
  char buf[snes_NAME_LEN];

  memset(buf, ' ', snes_NAME_LEN);
  strncpy(buf, rom->file, strlen(rom->file) > snes_NAME_LEN ?
          snes_NAME_LEN : strlen(rom->file));
  quickfwrite(buf, rom->header_start + rom->buheader_len + 16, snes_NAME_LEN,
              filebackup(rom->rom), "r+b");

  return 0;
}


int snes_chk(struct ucon64_ *rom)
{
  int image = rom->header_start + rom->buheader_len;

  if (snes_splitted!=0)
  {
    printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
    return -1;
  }

  filehexdump(rom->rom, image + 44, 4);
  filebackup(rom->rom);

  // change inverse checksum
  quickfputc(rom->rom, image + 44, (0xffff - rom->current_internal_crc), "r+b");      // low byte
  quickfputc(rom->rom, image + 45, (0xffff - rom->current_internal_crc) >> 8, "r+b"); // high byte
  // change checksum
  quickfputc(rom->rom, image + 46, rom->current_internal_crc, "r+b");      // low byte
  quickfputc(rom->rom, image + 47, rom->current_internal_crc >> 8, "r+b"); // high byte

  filehexdump(rom->rom, image + 44, 4);

  return 0;
}

#ifdef  BACKUP
int snes_xswc(struct ucon64_ *rom)
{
  if (access(rom->rom, F_OK) != 0)              // file does not exist -> dump cartridge
    swc_read_rom(rom->rom, rom->parport);
  else                                          // file exists -> send it to the copier
  {
    if (snes_splitted != 0)
    {
      printf("ERROR: this ROM seems to be splitted (use -ns to override)\n");
      return -1;
    }

    if (rom->buheader_len == 0)
    {
      printf("ERROR: this ROM has no header. Convert to an SWC compatible format (use -swc).\n");
      return -1;
    }

    if (rom->interleaved)
    {
      printf("ERROR: this ROM seems to be interleaved but the SWC doesn't support\n"
             "       interleaved ROMs (use -nint to override or -dint to convert)\n");
      return -1;
    }

    swc_write_rom(rom->rom, rom->parport, snes_sramsize);
  }
  printf("\n");
  return 0;
}

int snes_xswcs(struct ucon64_ *rom)
{
  if (access(rom->rom, F_OK) != 0)              // file does not exist -> dump SRAM contents
    swc_read_sram(rom->rom, rom->parport);
  else                                          // file exists -> restore SRAM
    swc_write_sram(rom->rom, rom->parport);

  printf("\n");
  return 0;
}
#endif


int testinterleaved(struct ucon64_ *rom)
{
  int interleaved = 0, map_type;
  char name[22];                                // NOT snes_NAME_LEN

  map_type = quickfgetc(rom->rom, rom->header_start + rom->buheader_len + 37);
  if (snes_hirom)
  {
    if ((map_type & 0xf0) == 0x20 || (map_type & 0xf0) == 0x30)
    {
      switch (map_type & 0xf)
      {
        case 0:
        case 3:
          interleaved = 1;
          break;
      }
    }
  }
  else
  {
    if ((map_type & 0xf0) == 0x20 || (map_type & 0xf0) == 0x30)
    {
      switch (map_type & 0xf)
      {
        case 1:
          if (strncmp(rom->name, "TREASURE HUNTER G", 17) != 0)
            interleaved = 1;
          break;
        case 5:
          interleaved = 1;
          rom_is_top = 1;
          break;
      }
    }

    quickfread(name, snes_ROM_HEADER_START + rom->buheader_len + 16, 22, rom->rom);
    if (strncmp(name, "YUYU NO QUIZ DE GO!GO!", 22) == 0)
      interleaved = 0;
    else if (strncmp(name, "SP MOMOTAROU DENTETSU2", 22) == 0)
      interleaved = 0;
  }

  return interleaved;
}

int snes_deinterleave(struct ucon64_ *rom, unsigned char *rom_buffer)
{
  unsigned char blocks[256], tmp[0x8000], b;
  int nblocks, i, j, size, org_hirom;
#ifdef  ALT_HILO
  int hi_score, lo_score;
#endif

  org_hirom = snes_hirom;
  size = rom->bytes - rom->buheader_len;
  nblocks = ((size / (8*1024))*8*1024) >> 16;

  if (0)
  /*
    "if (rom_is_top)" does not seem to work for interleaved ToP, but this does!
    This code comes from Snes9x, but for Snes9x applies the same...
  */
  {
    nblocks = 0x60;
    for (i = 0; i < 0x40; i += 2)
    {
      blocks[i + 0] = (i >> 1) + 0x20;
      blocks[i + 1] = (i >> 1) + 0x00;
    }
    for (i = 0; i < 0x80; i += 2)
    {
      blocks[i + 0x40] = (i >> 1) + 0x80;
      blocks[i + 0x41] = (i >> 1) + 0x40;
    }
    snes_hirom = snes_HIROM;
  }
  else if (rom->interleaved == 2)
  {
    for (i = 0; i < nblocks * 2; i++)
      blocks[i] = (i & ~0x1e) | ((i & 2) << 2) | ((i & 4) << 2) |
                  ((i & 8) >> 2) | ((i & 16) >> 2);
  }
  else
  {
    snes_hirom = snes_hirom ? 0 : snes_HIROM;
    for (i = 0; i < nblocks; i++)
    {
      blocks[i*2] = i + nblocks;
      blocks[i*2 + 1] = i;
    }
  }

  for (i = 0; i < nblocks * 2; i++)
  {
    for (j = i; j < nblocks * 2; j++)
    {
      if (blocks[j] == i)
      {
        memmove(tmp, &rom_buffer[blocks[j] * 0x8000], 0x8000);
        memmove(&rom_buffer[blocks[j] * 0x8000], &rom_buffer[blocks[i] * 0x8000], 0x8000);
        memmove(&rom_buffer[blocks[i] * 0x8000], tmp, 0x8000);
        b = blocks[j];
        blocks[j] = blocks[i];
        blocks[i] = b;
        break;
      }
    }
  }

#ifdef  ALT_HILO
  hi_score = scoreHiROM(rom_buffer, size);
  lo_score = scoreLoROM(rom_buffer, size);

  if (!force_interleaved &&
      ((snes_hirom && (lo_score >= hi_score || hi_score < 0)) ||
      (!snes_hirom && (hi_score > lo_score || lo_score < 0))))
  {                                             // ROM seems to be non-interleaved after all
    quickfread(rom_buffer, rom->buheader_len, size, rom->rom);
    rom->interleaved = 0;
  }
#endif

  return org_hirom;
}

long snes_init_(struct ucon64_ *rom)
{
  int x, image;
  long value = 0;

  image = snes_ROM_HEADER_START + snes_hirom + rom->buheader_len;
  for (x = 0; x < 4; x++)                       // don't use rom->header_start here!
  {
    value += (x==1 || x==3) ?
               (quickfgetc(rom->rom, image + 44 + x)*0x0100) :
               quickfgetc(rom->rom, image + 44 + x);
  }

  return value;
}


int snes_init(struct ucon64_ *rom)
{
  unsigned char *rom_buffer;
  int x, y, size, fsize, image, maker_byte;
  char buf[MAXBUFSIZE], *snes_MANUFACTURER[] = {
"0 Unknown Manufacturer",
"Nintendo",
"2",
"Imagineer-Zoom",
"4",
"Zamuse",
"Falcom",
"7",
"Capcom",
"HOT-B",
"Jaleco",
"Coconuts",
"Coconuts",
"13",
"Technos",
"Pack-In-Video",
"16",
"Starfish",
"Gremlin Graphics",
"Electronic Arts",
"20",
"Cobra Team",
"Human",
"KOEI",
"Hudson Soft",
"Yanoman",
"Yanoman",
"27",
"Tecmo Products",
"29",
"Open System",
"Virgin Games",
"KSS",
"Sunsoft",
"Pow",
"Micro World",
"36",
"37",
"Enix",
"Loriciel/Electro Brain",
"Kemco",
"Seta",
"42",
"43",
"44",
"Visit",
"46",
"47",
"Viacom New Media",
"Carrozzeria",
"Dynamic",
"?" /*0x33*/,
"Magifact",
"Hect",
"54",
"55",
"56",
"57",
"58",
"Arcade Zone",
"Empire Software",
"Loriciel",
"62",
"63",
"Seika",
"Ubi Soft",
"66",
"67",
"68",
"69",
"System 3",
"Spectrum HoloByte",
"72",
"Irem",
"74",
"Raya Systems",
"Renovation Products*",
"Malibu Games",
"78",
"U.S. Gold",
"Absolute",
"Acclaim",
"Activision",
"American Sammy",
"Gametek",
"Hi Tech Entertainment",
"LJN",
"87",
"88",
"89",
"Mindscape",
"91",
"92",
"Tradewest",
"94",
"American Softworks",
"Titus",
"Virgin Interactive Entertainment",
"Maxis*",
"Origin/FCI/Pony Canyon",
"100",
"101",
"102",
"Ocean",
"104",
"Electronic Arts",
"106",
"Laserbeam",
"108",
"109",
"Elite",
"Electro Brain",
"Infogrames",
"Interplay",
"LucasArts",
"Parker Brothers",
"Konami",
"STORM*",
"118",
"119",
"T*HQ",
"Accolade",
"Triffix",
"123",
"Microprose",
"125",
"126",
"Kemco",
"Misawa",
"Teichio",
"Namco",
"Lozc",
"KOEI",
"133",
"T.I.M.",
"135",
"Datam Polystar",
"137",
"138",
"Bullet-Proof Software",
"Vic Tokai",
"141",
"Character Soft",
"I'Max",
"Takara",
"Chun Soft",
"Video System",
"Bec",
"148",
"Varie",
"150",
"Kaneco",
"152",
"Pack-In-Video",
"Nichibutsu",
"Tecmo",
"Imagineer",
"157",
"158",
"159",
"Telenet",
"Hori",
"162",
"163",
"Konami",
"K. Amusement Leasing Co.",
"166",
"Takara",
"168",
"Technos",
"JVC",
"171",
"TOEI Animation",
"Toho",
"174",
"Namco",
"176",
"ASCII",
"Bandai",
"179",
"Enix",
"181",
"Halken",
"183",
"184",
"185",
"Culture Brain",
"Sunsoft",
"Toshiba EMI",
"Sony Imagesoft",
"190",
"Sammy",
"Taito",
"193",
"Kemco",
"Square",
"Tokuma Soft",
"Data East",
"Tonkinhouse",
"199",
"KOEI",
"201",
"Konami",
"NTVIC",
"204",
"Meldac",
"Pony Canyon",
"Sotsu Agency/Sunrise",
"Disco/Taito",
"Sofel",
"Quest",
"Sigma*",
"212",
"213",
"Naxat",
"215",
"Capcom",
"Banpresto",
"Tomy",
"Acclaim",
"220",
"NCS",
"Human",
"Altron",
"Jaleco",
"225",
"Yutaka",
"227",
"T&Esoft",
"Epoch",
"230",
"Athena",
"Asmik",
"Natsume",
"King Records",
"Atlus",
"Sony Music Entertainment",
"237",
"IGS",
"239",
"240",
"Motown Software",
"Left Field Entertainment",
"Beam Software/Extreme",
"TecMagik",
"245",
"246",
"247",
"248",
"Cybersoft",
"250",
"251",
"252",
"253",
"Davidson",
"Unknown Maker"
}, *snes_COUNTRY[] = {
"Japan",
"U.S.A.",
"Europe, Australia, Oceania and Asia",
"Sweden",
"Finland",
"Denmark",
"France",
"Holland",
"Spain",
"Germany, Austria and Switzerland",
"Italy",
"Hong Kong and China",
"Indonesia",
"Korea",
"14", "15", "16", "17", "18", "19",
"20", "21", "22", "23", "24", "25", "26", "27", "28", "29",
"30", "31", "32", "33", "34", "35", "36", "37", "38", "39",
"40", "41", "42", "43", "44", "45", "46", "47", "48", "49",
"50", "51", "52", "53", "54", "55", "56", "57", "58", "59",
"60", "61", "62", "63", "64", "65", "66", "67", "68", "69",
"70", "71", "72", "73", "74", "75", "76", "77", "78", "79",
"80", "81", "82", "83", "84", "85", "86", "87", "88", "89",
"90", "91", "92", "93", "94", "95", "96", "97", "98", "99",
"100", "101", "102", "103", "104", "105", "106", "107", "108", "109",
"110", "111", "112", "113", "114", "115", "116", "117", "118", "119",
"120", "121", "122", "123", "124", "125", "126", "127", "128", "129",
"130", "131", "132", "133", "134", "135", "136", "137", "138", "139",
"140", "141", "142", "143", "144", "145", "146", "147", "148", "149",
"150", "151", "152", "153", "154", "155", "156", "157", "158", "159",
"160", "161", "162", "163", "164", "165", "166", "167", "168", "169",
"170", "171", "172", "173", "174", "175", "176", "177", "178", "179",
"180", "181", "182", "183", "184", "185", "186", "187", "188", "189",
"190", "191", "192", "193", "194", "195", "196", "197", "198", "199",
"200", "201", "202", "203", "204", "205", "206", "207", "208", "209",
"210", "211", "212", "213", "214", "215", "216", "217", "218", "219",
"220", "221", "222", "223", "224", "225", "226", "227", "228", "229",
"230", "231", "232", "233", "234", "235", "236", "237", "238", "239",
"240", "241", "242", "243", "244", "245", "246", "247", "248", "249",
"250", "251", "252", "253", "254",
"Unknown Country"
}, *snes_ROMTYPE2[] = {
"+DSP1",
"+SuperFX",
"+OBC1",
"+SA-1",
"+S-DD1",
"+5",
"+6",
"+7",
"+8",
"+9",
"+10",
"+11",
"+12",
"+13",
"+14",
"+Custom"
}, *snes_ROMTYPE[] = {
"ROM only",
"ROM+RAM",
"ROM+Save RAM",
"ROM and DSP1 chip",
"ROM+RAM+DSP1 chip",
"ROM+Save RAM+DSP1 chip",
"6", "7", "8", "9",
"10", "11", "12", "13", "14", "15", "16", "17", "18",
"ROM and Super FX chip",
"20", "21", "22", "23", "24", "25", "26", "27", "28", "29",
"30", "31", "32", "33", "34", "35", "36", "37", "38", "39",
"40", "41", "42", "43", "44", "45", "46", "47", "48", "49",
"50", "51", "52",
"ROM, RAM, Save RAM and SA-1",
"54", "55", "56", "57", "58", "59",
"60", "61", "62", "63", "64", "65", "66", "67", "68", "69",
"70", "71", "72", "73", "74", "75", "76", "77", "78", "79",
"80", "81", "82", "83", "84", "85", "86", "87", "88", "89",
"90", "91", "92", "93", "94", "95", "96", "97", "98", "99",
"100", "101", "102", "103", "104", "105", "106", "107", "108", "109",
"110", "111", "112", "113", "114", "115", "116", "117", "118", "119",
"120", "121", "122", "123", "124", "125", "126", "127", "128", "129",
"130", "131", "132", "133", "134", "135", "136", "137", "138", "139",
"140", "141", "142", "143", "144", "145", "146", "147", "148", "149",
"150", "151", "152", "153", "154", "155", "156", "157", "158", "159",
"160", "161", "162", "163", "164", "165", "166", "167", "168", "169",
"170", "171", "172", "173", "174", "175", "176", "177", "178", "179",
"180", "181", "182", "183", "184", "185", "186", "187", "188", "189",
"190", "191", "192", "193", "194", "195", "196", "197", "198", "199",
"200", "201", "202", "203", "204", "205", "206", "207", "208", "209",
"210", "211", "212", "213", "214", "215", "216", "217", "218", "219",
"220", "221", "222", "223", "224", "225", "226",
"ROM, RAM and GameBoy data",
"228", "229", "230", "231", "232", "233", "234","235","236","237","238","239",
"240", "241", "242", "243", "244", "245",
"ROM and DSP2 chip",
"247", "248", "249", "250", "251", "252", "253", "254", ""
};

  rom_is_top = 0;                               // reset this var, for -lsv

  rom->console = ucon64_SNES;
  rom->bytes = quickftell(rom->rom);            // here -> right value for SRAM files

  rom->misc[0] = 0;

  if (quickfgetc(rom->rom, 8) == 0xaa && quickfgetc(rom->rom, 9) == 0xbb &&
      quickfgetc(rom->rom, 10) == 5)
  {
    rom->buheader_len = smc_HEADER_LEN;
    strcpy(rom->name, "Name: N/A");
    strcpy(rom->title, "Super Wild Card SRAM file\n");
    strcpy(rom->copier, swc_TITLE);
    strcpy(rom->manufacturer, "Manufacturer: You?");
    strcpy(rom->country, "Country: Your country?");
    rom->has_internal_crc = 0;
    return 0;                                   // rest is nonsense for SRAM file
  }

  // the order in which the different combinations of values for snes_hirom and
  // rom->buheader_len are made is important, don't change it without a good reason!
  snes_hirom = 0;
  rom->buheader_len = 0;
  if (snes_init_(rom) != 0xffff)
  {
    snes_hirom = 0;
    rom->buheader_len = smc_HEADER_LEN;
    if (snes_init_(rom) != 0xffff)
    {
      snes_hirom = snes_HIROM;
      rom->buheader_len = 0;
      if (snes_init_(rom) != 0xffff)
      {
        snes_hirom = snes_HIROM;
        rom->buheader_len = smc_HEADER_LEN;
        if (snes_init_(rom) != 0xffff)
        {
          if (!argcmp(rom->argc, rom->argv, "-snes"))
          {
            rom->console = ucon64_UNKNOWN;
            return -1;
          }
        }
      }
    }
  }

  if (quickfgetc(rom->rom, 8) == 0xaa && quickfgetc(rom->rom, 9) == 0xbb &&
      quickfgetc(rom->rom, 10) == 4)
    rom->buheader_len = smc_HEADER_LEN;
  else
  {                                             // no SWC header -> alt. header check
    fsize = ((quickfgetc(rom->rom, 1) << 8) + quickfgetc(rom->rom, 0)) * 8*1024;
    fsize += smc_HEADER_LEN;                    // if correct header -> hdr[1] high byte,
    if ((unsigned) fsize == rom->bytes)         //  hdr[0] low byte of # 8KB blocks in ROM
      rom->buheader_len = smc_HEADER_LEN;
  }

/*
  testinterleaved() needs the correct value for snes_hirom and rom->header_start.
  snes_hirom may be used only after the check for -hi/-nhi has been made. If ALT_HILO
  is defined however, rom->buheader_len must have the correct value in order to
  determine the value for snes_hirom. This can only be known after the check for
  -hd/-nhd. So, the order must be
  1. - check for -hd/-nhd
  2. - snes_hirom
  3. - check for -hi/-nhi
  4. - testinterleaved()
*/
  snes_splitted = testsplit(rom->rom);
  if (argcmp(rom->argc, rom->argv, "-ns"))
    snes_splitted = 0;
  if (argcmp(rom->argc, rom->argv, "-hd"))
    rom->buheader_len = smc_HEADER_LEN;
  if (argcmp(rom->argc, rom->argv, "-nhd"))
    rom->buheader_len = 0;

  size = rom->bytes - rom->buheader_len;
  if (size < 32768)
    if (!argcmp(rom->argc, rom->argv, "-snes"))
    {
      rom->console = ucon64_UNKNOWN;
      return -1;
    }

  if (!(rom_buffer = (unsigned char *) malloc(size)))
  {
    fprintf(STDERR, "Not enough memory for ROM buffer (%d bytes)\n", size);
    return -1;                                  // don't exit(), we might've been
  }                                             //  called from -lsv
  quickfread(rom_buffer, rom->buheader_len, size, rom->rom);

#ifdef  ALT_HILO
  snes_hirom = scoreHiROM(rom_buffer, size) > scoreLoROM(rom_buffer, size) ? snes_HIROM : 0;
#endif
  if (argcmp(rom->argc, rom->argv, "-hi"))
    snes_hirom = snes_HIROM;
  if (argcmp(rom->argc, rom->argv, "-nhi"))
    snes_hirom = 0;

  image = snes_ROM_HEADER_START + snes_hirom;
  rom->header_start = image;
  rom->header_len = snes_ROM_HEADER_LEN;

  rom->interleaved = testinterleaved(rom);
  if (argcmp(rom->argc, rom->argv, "-int"))
  {
    rom->interleaved = 1;
    force_interleaved = 1;                      // uCON64 needs to know if it was forced
  }                                             //  for snes_deinterleave()
  if (argcmp(rom->argc, rom->argv, "-int2"))
  {
    rom->interleaved = 2;
    force_interleaved = 2;
  }
  if (argcmp(rom->argc, rom->argv, "-nint"))
    rom->interleaved = 0;

  rom->buheader_start = 0;
  if (rom->buheader_len == 0)
    memset(rom->buheader, 0, smc_HEADER_LEN);   // better have this initialized
  else
    quickfread(rom->buheader, 0, rom->buheader_len, rom->rom);

  memcpy(rom->header, rom_buffer + rom->header_start, rom->header_len);

  memcpy(rom->name, rom_buffer + image + 16, snes_NAME_LEN);
  rom->name[snes_NAME_LEN] = 0;                 // terminate string

  strcpy(rom->title, snes_TITLE);
  if (!rom->buheader_len)
    strcpy(rom->copier, mgd_TITLE);
  else
  {
    if (!snes_hirom)                            // TODO rom->buheader checks instead of extcmp()
      sprintf(rom->copier, "%s", (!extcmp(rom->rom, "fig")) ?
                                   fig_TITLE :
                                   ((!extcmp(rom->rom, "swc")) ? swc_TITLE : smc_TITLE));
    else
      sprintf(rom->copier, "%s", (!extcmp(rom->rom, "fig")) ? fig_TITLE : swc_TITLE);
  }

  maker_byte = rom_buffer[image + 42];
  if (maker_byte == 0x33)
  {
    x = rom_buffer[image];
    y = rom_buffer[image + 1];

    strcpy(rom->manufacturer,
      snes_MANUFACTURER[(x < 0x3a ? (x - 0x30)*16 : (x - 0x37)*16)
                                    + (y < 0x3a ? y - 0x30 : y - 0x37)]);
  }
  else
  {
    if (maker_byte <= 0)
      strcpy(rom->manufacturer, "Demo or Beta ROM?");
    else
      strcpy(rom->manufacturer, snes_MANUFACTURER[maker_byte&0xff]);
  }

  strcpy(rom->country, snes_COUNTRY[rom_buffer[image + 41]&0xff]);

  sprintf(buf, "Internal Size: %.4f Mb\n", (float) (1 << (rom_buffer[image + 39] - 7)));
  strcat(rom->misc, buf);

  x = rom_buffer[image + 38];
  sprintf(buf, "RomType: (%x) %s", x, snes_ROMTYPE[x%3]);
  strcat(rom->misc, buf);
  sprintf(buf, "%s\n", (x & 0xf) >= 3 ? snes_ROMTYPE2[(x & 0xf0) >> 4] : "");
  strcat(rom->misc, buf);

  sprintf(buf, "RomSpeed: %s\n", rom_buffer[image + 37] & 0x10 ? "120ns (FastROM)" : "200ns (SlowROM)");
  strcat(rom->misc, buf);

  sprintf(buf, "HiROM: %s\n", !snes_hirom ? "No (<16Mb and/or before mid-1993)" : "Yes (>=16 Mb and/or after mid-1993)");
  strcat(rom->misc, buf);

  rom->has_internal_crc = 1;
  rom->internal_crc_len = rom->internal_crc2_len = 2;
  rom->current_internal_crc = snes_chksum(rom, rom_buffer);

  rom->internal_crc = rom_buffer[image + 46];           // low byte of checksum
  rom->internal_crc += rom_buffer[image + 47] << 8;     // high byte of checksum
  x = rom_buffer[image + 44];                           // low byte of inverse checksum
  x += rom_buffer[image + 45] << 8;                     // high byte of inverse checksum
  sprintf(buf, "Inverse checksum: %%s, 0x%%0%dlx + 0x%%0%dlx = 0x%%0%dlx %%s",
    rom->internal_crc2_len*2, rom->internal_crc2_len*2, rom->internal_crc2_len*2);
  sprintf(rom->internal_crc2, buf,
          (rom->current_internal_crc + x == 0xffff) ? "ok" : "bad",
          rom->current_internal_crc, x, rom->current_internal_crc + x,
          (rom->current_internal_crc + x == 0xffff) ? "" : "~0xffff");

  // do the following after call to snes_chksum() which might call snes_deinterleave()
  // which might change value of rom->interleaved
  sprintf(buf, "Interleaved: %s\n", rom->interleaved ? "Yes" : "No");
  strcat(rom->misc, buf);

  y = rom_buffer[image + 40];
  snes_sramsize = y ? (1 << (y + 3)) * 128 : 0;
  if (!snes_sramsize)
    sprintf(buf, "Save RAM: No\n");
  else
    sprintf(buf, "Save RAM: Yes, %d kBytes\n", snes_sramsize/1024);
  strcat(rom->misc, buf);
  sprintf(buf, "Version: 1.%d", rom_buffer[image + 43]);
  strcat(rom->misc, buf);

  free(rom_buffer);
  return 0;
}

int snes_special_bs(struct ucon64_ *rom, unsigned char *rom_buffer)
{
  int image = rom->header_start;
  unsigned char *game_name = rom_buffer + image + 16;

  if (!strncmp("BS DRAGON QUEST", (char *) game_name, 15))
    return 1;
  if (!strncmp("BUSTERS BS", (char *) game_name, 10))
    return 1;
  if (!strncmp("Mario Excite Bike", (char *) game_name, 17))
    return 1;
  if (!strncmp("BS_Dr Mario", (char *) game_name, 11))
    return 1;
  if (!strncmp("BS_SuperFam", (char *) game_name, 11))
    return 1;
  if (!strncmp("BS_Legend", (char *) game_name, 9))
    return 1;
  if (!strncmp("KARBYCOL", (char *) rom_buffer + image + 7, 8))
    return 1;
  if (game_name[0] == 0xdc && game_name[1] == 0xb2 && game_name[5] == 0xaa)
    return 1;                                   // bs wai wai
  if (game_name[0] == 0x8c && game_name[1] == 0x8e && game_name[5] == 0xb2)
    return 1;                                   // bs coin deck
  if (game_name[0] == 0x82 && game_name[1] == 0xb3 && game_name[5] == 0xaa)
    return 1;                                   // bs same game koma data
  if (game_name[0] == 0xbb && game_name[1] == 0xc3 && game_name[5] == 0x2d)
    return 1;                                   // ?, not bs zelda 3 (checked
                                                //  remix, 1 & 3)
  return 0;
}

int snes_check_bs(struct ucon64_ *rom, unsigned char *rom_buffer)
{
  unsigned char byte;
  int i, image = rom->header_start;

  if (snes_special_bs(rom, rom_buffer))
    return 1;
  if (rom_buffer[image + 0x24])
    return 0;
  byte = rom_buffer[image + 0x2a];              // makerid
  if (byte != 0x33 && byte != 0 && byte != 0xff )
    return 0;
  byte = rom_buffer[image + 0x28];              // rom makeup
  if (byte & 0x8e)
    return 0;
  if (!(byte & 0x70))
    return 0;
  byte = rom_buffer[image + 0x20];
  if (byte != 0x01 && byte != 0x03 && byte != 0x07 && byte != 0x0f && byte != 0xff)
    return 0;
//  if (byte != 0x00 && byte != 0x01 && byte != 0x20 && byte != 0x21 && byte != 0x30 && byte != 0x31)
//    return 0;
  byte = rom_buffer[image + 0x29];              // bs size;
  if (byte != 0x00 && byte != 0x10 && byte != 0x20 && byte != 0x30 && byte != 0x80)
    return 0;
//  if (!isalnum(rom_buffer[image + [0])) return 0;
//  if (!isalnum(rom_buffer[image + [1])) return 0;
  for (i = 16; i < 12 + 16 ; i++)
    if (rom_buffer[image + i] == 0xff)
      return 0;
  byte = rom_buffer[image + 0x26];              // bs month
  if (byte != 0x0 && byte != 0xff)
  {
    if (byte & 0x0f)
      return 0;
    if ((byte >> 4) > 12)
      return 0;
  }

  return 1;
}

int snes_chksum(struct ucon64_ *rom, unsigned char *rom_buffer)
{
  int i, power2 = 0, size, remainder, sum1 = 0, sum2 = 0;

  if (rom->interleaved)                         // keep original value for snes_hirom
    snes_hirom = snes_deinterleave(rom, rom_buffer);

  size = rom->bytes - rom->buheader_len;
  while (size >>= 1)
    power2++;
  size = 1 << power2;

  if (size >= MBIT && snes_check_bs(rom, rom_buffer))
  {                                             // Broadcast Satellaview ROM
    for (i = 0; i < rom->header_start; i++)
      sum1 += rom_buffer[i];
    for (i = rom->header_start + snes_ROM_HEADER_LEN; i < size; i++)
      sum1 += rom_buffer[i];
  }
  else
    for (i = 0; i < size; i++)                  // normal ROM
      sum1 += rom_buffer[i];

  remainder = (rom->bytes - rom->buheader_len) - size;
  for (i = 0; i < remainder; i++)
    sum2 += rom_buffer[size + i];

  if (remainder)
    sum1 += sum2*(size/remainder);

  if (!(!force_interleaved && rom->interleaved == 0))
    quickfread(rom_buffer, rom->buheader_len, rom->bytes - rom->buheader_len, rom->rom);

  return sum1 & 0xffff;
}

#define encodeSNES(x, y) transposed |= (((address & (0xc00000 >> (2*y))) << (2*y)) >> (2*x))
#define decodeSNES(x, y) address |= (((transposed & (0xc00000 >> (2*x))) << (2*x)) >> (2*y))

char mapSnesChar(char hex) {
  switch(toupper(hex)) {
    case '0': return 'D';
    case '1': return 'F';
    case '2': return '4';
    case '3': return '7';
    case '4': return '0';
    case '5': return '9';
    case '6': return '1';
    case '7': return '5';
    case '8': return '6';
    case '9': return 'B';
    case 'A': return 'C';
    case 'B': return '8';
    case 'C': return 'A';
    case 'D': return '2';
    case 'E': return '3';
    case 'F': return 'E';
    default: break;
  }
  return ' ';
}

char unmapSnesChar(char genie) {
  switch(toupper(genie)) {
    case 'D': return '0';
    case 'F': return '1';
    case '4': return '2';
    case '7': return '3';
    case '0': return '4';
    case '9': return '5';
    case '1': return '6';
    case '5': return '7';
    case '6': return '8';
    case 'B': return '9';
    case 'C': return 'A';
    case '8': return 'B';
    case 'A': return 'C';
    case '2': return 'D';
    case '3': return 'E';
    case 'E': return 'F';
    default: break;
  }
  return ' ';
}

int
gameGenieDecodeSNES(const char *in, char *out) {
  int value;
  long address;
  long transposed;

  if (!isxdigit((int) in[0]) || !isxdigit((int) in[1]) ||
      !isxdigit((int) in[2]) || !isxdigit((int) in[3]) ||
      in[4] != '-' ||
      !isxdigit((int) in[5]) || !isxdigit((int) in[6]) ||
      !isxdigit((int) in[7]) || !isxdigit((int) in[8]) ||
      in[9] != 0)
    return -1;

  value = hexByteValue(unmapSnesChar(in[0]), unmapSnesChar(in[1]));

  transposed = hexValue(unmapSnesChar(in[8])) +
    (hexValue(unmapSnesChar(in[7])) << 4) +
    (hexValue(unmapSnesChar(in[6])) << 8) +
    (hexValue(unmapSnesChar(in[5])) << 12) +
    (hexValue(unmapSnesChar(in[3])) << 16) +
    (hexValue(unmapSnesChar(in[2])) << 20);

  address = 0;
  decodeSNES(0, 4);
  decodeSNES(1, 5);
  decodeSNES(2, 8);
  decodeSNES(3, 9);
  decodeSNES(4, 7);
  decodeSNES(5, 0); 
  decodeSNES(6, 1);
  decodeSNES(7, 10);
  decodeSNES(8, 11); 
  decodeSNES(9, 2);
  decodeSNES(10, 3);
  decodeSNES(11, 6);

  sprintf(out, "%06lX:%02X", address, value);

  return 0;
}

int
gameGenieEncodeSNES(const char *in, char *out) {
  int value;
  long address;
  long transposed;

  if (!isxdigit((int) in[0]) || !isxdigit((int) in[1]) ||
      !isxdigit((int) in[2]) || !isxdigit((int) in[3]) ||
      !isxdigit((int) in[4]) || !isxdigit((int) in[5]) ||
      in[6] != ':' ||
      !isxdigit((int) in[7]) || !isxdigit((int) in[8]) ||
      in[9] != 0)
    return -1;

  value = hexByteValue(mapSnesChar(in[7]), mapSnesChar(in[8]));
  sscanf(in, "%lx", &address);

  transposed = 0;
  encodeSNES(0, 4);
  encodeSNES(1, 5);
  encodeSNES(2, 8);
  encodeSNES(3, 9);
  encodeSNES(4, 7);
  encodeSNES(5, 0);
  encodeSNES(6, 1);
  encodeSNES(7, 10);
  encodeSNES(8, 11);
  encodeSNES(9, 2);
  encodeSNES(10, 3);
  encodeSNES(11, 6);

  sprintf(out, "%02X%c%c-%c%c%c%c", value,
          mapSnesChar(hexDigit((transposed & 0xf00000) >> 20)),
          mapSnesChar(hexDigit((transposed & 0x0f0000) >> 16)),
          mapSnesChar(hexDigit((transposed & 0x00f000) >> 12)),
          mapSnesChar(hexDigit((transposed & 0x000f00) >> 8)),
          mapSnesChar(hexDigit((transposed & 0x0000f0) >> 4)),
          mapSnesChar(hexDigit((transposed & 0x00000f))));

  return 0;
}



#ifdef  ALT_HILO
int scoreHiROM(unsigned char *rom_buffer, int rom_size)
{
  int score = 0;

  if ((rom_buffer[0xff00 + 0xdc] + (rom_buffer[0xff00 + 0xdd] << 8) +
       rom_buffer[0xff00 + 0xde] + (rom_buffer[0xff00 + 0xdf] << 8)) == 0xffff)
    score += 2;

  if (rom_buffer[0xff00 + 0xda] == 0x33)
    score += 2;
  if ((rom_buffer[0xff00 + 0xd5] & 0xf) < 4)
    score += 2;
  if (!(rom_buffer[0xff00 + 0xfd] & 0x80))
    score -= 4;
#if 0
  // this seems to work by accident for Snes9x (letting rom_size (CalculatedSize)
  // be 0 for first call)
  if (rom_size > 1024 * 1024 * 3)
    score += 4;
#endif
  if ((1 << (rom_buffer[0xff00 + 0xd7] - 7)) > 48)
    score -= 1;
  if (!allascii(rom_buffer + 0xff00 + 0xb0, 6))
    score -= 1;
  if (!allascii(rom_buffer + 0xff00 + 0xc0, snes_NAME_LEN))
    score -= 1;

  return score;
}

int scoreLoROM(unsigned char *rom_buffer, int rom_size)
{
  int score = 0;

  if ((rom_buffer[0x7f00 + 0xdc] + (rom_buffer[0x7f00 + 0xdd] << 8) +
       rom_buffer[0x7f00 + 0xde] + (rom_buffer[0x7f00 + 0xdf] << 8)) == 0xffff)
    score += 2;

  if (rom_buffer[0x7f00 + 0xda] == 0x33)
    score += 2;
  if ((rom_buffer[0x7f00 + 0xd5] & 0xf) < 4)
    score += 2;
#if 1
  // this also seems strange, ROMs are always less than 16MB
  if (rom_size <= 1024 * 1024 * 16)
    score += 2;
#endif
  if (!(rom_buffer[0x7f00 + 0xfd] & 0x80))
    score -= 4;
  if ((1 << (rom_buffer[0x7f00 + 0xd7] - 7)) > 48)
    score -= 1;
  if (!allascii(rom_buffer + 0x7f00 + 0xb0, 6))
    score -= 1;
  if (!allascii(rom_buffer + 0x7f00 + 0xc0, snes_NAME_LEN))
    score -= 1;

  return score;
}
#endif


int snes_usage(int argc,char *argv[])
{

printf("%s\n"
       "  -snes         force recognition\n"
       "  -hi           force ROM is HiROM\n"
       "  -nhi          force ROM is not HiROM\n"
       "  -hd           force ROM has SMC/FIG/SWC header (+512 Bytes)\n"
       "  -nhd          force ROM has no SMC/FIG/SWC header (MGD2/MGH/RAW)\n"
       "  -ns           force ROM is not splitted\n"
       "  -int          force ROM is in interleaved format\n"
       "  -int2         force ROM is in interleaved format 2\n"
       "  -nint         force ROM is not in interleaved format\n"
       "  *             show info (default)\n"
       "  -n            change ROM name; $FILE=NEWNAME\n"
       "  -smc          convert to Super Magicom/SMC\n"
       "  -fig          convert to *Pro Fighter*/(all)FIG\n"
       "  -figs         convert Snes9x/ZSNES *.srm (SRAM) to *Pro Fighter*/(all)FIG\n"
       "  -swc          convert to Super Wild Card*/(all)SWC\n"
       "  -swcs         convert Snes9x/ZSNES *.srm (SRAM) to Super Wild Card*/(all)SWC\n"
       "  -mgd          convert to Multi Game*/MGD2/MGH/RAW\n"
       "  -gd3          convert to Professor SF(2) Game Doctor SF3/6/7\n"
       "  -ufos         convert Snes9x/ZSNES *.srm (SRAM) to Super UFO\n"
//       "  -gdf          fix Professor SF(2) Game Doctor SF3/6/7 savegame problems\n"
       "  -dint         convert ROM to non-interleaved format\n"
//       "TODO:  -int          convert ROM to interleaved format\n"
       "  -col          convert 0xRRGGBB (html) <-> 0xXXXX (snes); $ROM=0xCOLOR\n"
       "                this routine was used to find green colors in games and\n"
       "                to replace them with red colors (blood mode)\n"
//       "TODO:  -sx      convert to Snes9X (Emulator)/S9X save state; $ROM=SAVESTATE\n"
//       "TODO:  -zs      convert to ZSNES (Emulator) save state; $ROM=SAVESTATE\n"
//       "TODO:  -xzs     extract GFX from ZSNES (Emulator) save state; $ROM=SAVESTATE\n"
//       "TODO:  -spc     convert SPC sound to WAV; $ROM=SPCFILE\n"
       "  -j            join splitted ROM\n"
       "  -s            split ROM into 4Mb parts (for backup unit(s) with fdd)\n"
       "  -p            pad ROM to full Mb\n"
       "  -k            remove protection (crack)\n"
       "  -f            remove NTSC/PAL protection\n"
       "  -l            convert to SlowROM\n"
       "  -chk          fix ROM checksum\n"
       "  -gge          encode GameGenie code; $ROM=AAAAAA:VV\n"
       "  -ggd          decode GameGenie code; $ROM=XXXX-XXXX\n"
       "  -gg           apply GameGenie code (permanent); $FILE=XXXX-XXXX\n",
       snes_TITLE);

bsl_usage(argc,argv);
ips_usage(argc,argv);

#ifdef  BACKUP
swc_usage(argc,argv);
fig_usage(argc,argv);
smc_usage(argc,argv);
mgd_usage(argc,argv);
#endif

printf("\n");


return 0;
}
